from rbc.tests import heavydb_fixture
import pytest

rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), load_test_data=False):
        define(o)
        yield o


def define(heavydb):
    # Defines simple methods just to verify if NRT is working and if any there
    # are any symbol
    #
    # List of methods:
    # [X] set.add
    # [X] set.clear
    # [X] set.copy
    # [X] set.difference
    # [X] set.difference_update
    # [X] set.discard
    # [X] set.intersection
    # [X] set.intersection_update
    # [X] set.isdisjoint
    # [X] set.issubset
    # [X] set.issuperset
    # [X] set.pop
    # [X] set.remove
    # [X] set.symmetric_difference
    # [X] set.symmetric_difference_update
    # [ ] set.union
    # [X] set.update

    @heavydb("i32(i32, TextEncodingNone)", devices=['cpu'])
    def test_set(t, method):
        method_str = method.to_string()
        s = set([1, 2, 3, 4, 5])
        if method_str == 'add':
            return len(s)
        elif method_str == 'clear':
            s.clear()
            return len(s) == 0
        elif method_str == 'copy':
            s2 = s.copy()
            return len(s) == len(s2) == 5
        elif method_str == 'difference':
            s2 = set([1, 3, 5])
            return s.difference(s2) == {2, 4}
        elif method_str == 'difference_update':
            s.difference_update({1, 2, 3})
            return len(s) == 2
        elif method_str == 'discard':
            s.discard(1)
            return len(s) == 4
        elif method_str == 'intersection':
            return s.intersection({1, 3, 5}) == {1, 3, 5}
        elif method_str == 'intersection_update':
            s.intersection_update({1, 2})
            return len(s) == 2
        elif method_str == 'isdisjoint':
            return s.isdisjoint({0, 6})
        elif method_str == 'issubset':
            return s.issubset({0, 1, 2, 3, 4, 5})
        elif method_str == 'issuperset':
            return s.issuperset({1, 2, 3})
        elif method_str == 'pop':
            s.pop()
            return len(s) == 4
        elif method_str == 'remove':
            s.remove(2)
            return len(s) == 4
        elif method_str == 'symmetric_difference':
            s2 = {1, 2}
            return len(s.symmetric_difference(s2)) == 3
        elif method_str == 'symmetric_difference_update':
            s.symmetric_difference_update({0, 6})
            return len(s) == 7
        # elif method_str == 'union':
        #     return len(s.union({0})) == 6
        elif method_str == 'update':
            s.update({0})
            return len(s) == 6
        return 0


@pytest.mark.parametrize('method,ans', [('add', 5), ('clear', 1), ('copy', 1),
                                        ('difference', 1), ('difference_update', 1),
                                        ('discard', 1), ('intersection', 1),
                                        ('intersection_update', 1), ('isdisjoint', 1),
                                        ('issubset', 1), ('issuperset', 1), ('pop', 1),
                                        ('remove', 1), ('symmetric_difference', 1),
                                        ('symmetric_difference_update', 1),
                                        ('union', 1), ('update', 1)])
def test_set_methods(heavydb, method, ans):
    skip_list = ('union',)
    if method in skip_list:
        pytest.skip(method)
    _, result = heavydb.sql_execute(f"select test_set(1, '{method}');")
    assert list(result)[0] == (ans,)



import pytest
import numpy as np

from rbc.stdlib import array_api
from rbc.typesystem import Type
from rbc.heavydb import type_to_type_name
from numba.core import types


rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def heavydb():
    # TODO: use heavydb_fixture from rbc/tests/__init__.py
    config = rbc_heavydb.get_client_config(debug=not True)
    m = rbc_heavydb.RemoteHeavyDB(**config)
    table_name = 'rbc_test_heavydb_array'

    m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')
    sqltypes = ['FLOAT[]', 'DOUBLE[]',
                'TINYINT[]', 'SMALLINT[]', 'INT[]', 'BIGINT[]',
                'BOOLEAN[]']
    # todo: TEXT ENCODING DICT, TEXT ENCODING NONE, TIMESTAMP, TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://docs.heavy.ai/sql/data-definition-ddl/datatypes-and-fixed-encoding
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(
        'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});'
        .format(**locals()))

    def row_value(row, col, colname):
        if colname == 'b':
            return 'ARRAY[%s]' % (', '.join(
                ("'true'" if i % 2 == 0 else "'false'")
                for i in range(-3, 3)))
        if colname.startswith('f'):
            return 'ARRAY[%s]' % (', '.join(
                str(row * 10 + i + 0.5) for i in range(-3, 3)))
        return 'ARRAY[%s]' % (', '.join(
            str(row * 10 + i) for i in range(-3, 3)))

    rows = 5
    for i in range(rows):
        table_row = ', '.join(str(row_value(i, j, n))
                              for j, n in enumerate(colnames))
        m.sql_execute(f'INSERT INTO {table_name} VALUES ({table_row})')
    m.table_name = table_name
    yield m
    try:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')
    except Exception as msg:
        print('%s in deardown' % (type(msg)))


def np_ones(sz):
    return array_api.ones(sz, dtype=types.int32)


def np_ones_dtype(sz):
    return array_api.ones(sz)


def np_ones_like_dtype(i4):
    return array_api.ones_like(i4, dtype=types.double)


def np_ones_like(i4):
    return array_api.ones_like(i4)


def np_empty(sz):
    return array_api.empty(sz, np.int32)


def np_empty_dtype(sz):
    return array_api.empty(sz)


def np_empty_like(i4):
    return array_api.empty_like(i4)


def np_zeros(sz):
    return array_api.zeros(sz, np.int32)


def np_zeros_dtype(sz):
    return array_api.zeros(sz)


def np_zeros_like(i4):
    return array_api.zeros_like(i4)


def np_zeros_like_dtype(i4):
    return array_api.zeros_like(i4, dtype=types.double)


def np_full(sz, fill_value):
    return array_api.full(sz, fill_value, dtype=types.double)


def np_full_dtype(sz, fill_value):
    return array_api.full(sz, fill_value)


def np_full_like(i1, fill_value):
    return array_api.full_like(i1, fill_value)


def np_full_like_dtype(i1, fill_value):
    return array_api.full_like(i1, fill_value, dtype=types.double)


def np_cumsum(sz):
    a = array_api.ones(sz)
    return array_api.cumsum(a)


array_methods = [
    ('full', 'double[](int64, double)', (5, 3), np.full(5, 3, dtype=np.int32)),
    ('full_dtype', 'double[](int64, double)', (5, 3), np.full(5, 3)),
    ('full_like', 'int8[](int8[], int32)', ('i1', 3), np.full(6, 3, dtype='b')),  # noqa: E501
    ('full_like_dtype', 'double[](int8[], double)', ('i1', 3.0), np.full(6, 3, dtype='q')),  # noqa: E501
    ('ones', 'int32[](int64)', (5,), np.ones(5, dtype=np.int32)),
    ('ones_dtype', 'double[](int64)', (5,), np.ones(5)),
    ('ones_like', 'int32[](int32[])', ('i4',), np.ones(6, dtype='i')),
    ('ones_like_dtype', 'double[](int32[])', ('i4',), np.ones(6, dtype='q')),
    ('zeros', 'int32[](int64)', (5,), np.zeros(5, dtype=np.int32)),
    ('zeros_like', 'int32[](int32[])', ('i4',), np.zeros(6, dtype='i')),
    ('zeros_like_dtype', 'double[](int32[])', ('i4',), np.zeros(6, dtype='q')),
    ('zeros_dtype', 'double[](int64)', (5,), np.zeros(5)),
    ('empty', 'int32[](int64)', (5,), np.empty(5, dtype=np.int32)),
    ('empty_like', 'int8[](int8[])', ('i1',), np.empty(6, dtype='i')),
    ('empty_like', 'int32[](int32[])', ('i4',), np.empty(6, dtype='i')),
    ('empty_like', 'float[](float[])', ('f4',), np.empty(6, dtype='f')),
    ('empty_like', 'double[](double[])', ('f8',), np.empty(6, dtype='f')),
    ('empty_dtype', 'double[](int64)', (5,), np.empty(5)),
    ('cumsum', 'double[](int32)', (5,), np.arange(1, 6)),
]


@pytest.mark.parametrize("method, signature, args, expected", array_methods,
                         ids=[item[0] for item in array_methods])
def test_array_methods(heavydb, method, signature, args, expected):
    heavydb.reset()

    fn = heavydb(signature)(eval('np_{}'.format(method)))

    query = 'select np_{method}'.format(**locals()) + \
            '(' + ', '.join(map(str, args)) + ')' + \
            ' from {heavydb.table_name};'.format(**locals())

    _, result = heavydb.sql_execute(query)
    out = list(result)[0]

    if 'empty' in method:
        assert out == ([None] * len(expected),)
    else:
        assert np.array_equal(expected, out[0]), 'np_' + method


@pytest.mark.parametrize('col', ('i4', 'i8', 'f4'))
def test_dtype_attribute(heavydb, col):
    heavydb.reset()

    @heavydb('T[](T[])', T=['int32', 'int64', 'float32'], devices=['cpu'])
    def zeros(x):
        z = array_api.zeros(len(x), x.dtype)
        return z

    query = f'select zeros({col}) from {heavydb.table_name} limit 1;'
    _, result = heavydb.sql_execute(query)
    assert np.all(list(result)[0][0] == np.zeros(6, dtype=col))


dtype_to_col = dict(int8_t='i1', int16_t='i2', int32_t='i4',
                    int64_t='i8', float='f4', double='f8')
functions = ('zeros', 'zeros_like',
             'full', 'full_like',
             'empty', 'empty_like',
             'ones', 'ones_like',)


@pytest.mark.parametrize('dtype', ('int32_t', 'float'))  # test all dtypes?
@pytest.mark.parametrize('func_name', functions)
def test_literal_string_dtype(heavydb, dtype, func_name):
    heavydb.reset()

    func = getattr(array_api, func_name)

    if func_name == 'full_like':
        fill_value = 0.0 if dtype in ('float', 'double') else 0

        @heavydb('T[](T[])', T=[dtype], devices=['cpu'])
        def fn(arr):
            return func(arr, fill_value, dtype=dtype)
        query = f'select fn({dtype_to_col[dtype]}) from {heavydb.table_name} limit 1;'
    elif func_name.endswith('_like'):

        @heavydb('T[](T[])', T=[dtype], devices=['cpu'])
        def fn(arr):
            return func(arr, dtype=dtype)
        query = f'select fn({dtype_to_col[dtype]}) from {heavydb.table_name} limit 1;'
    elif func_name == 'full':

        @heavydb('T[](int32, T)', T=[dtype], devices=['cpu'])
        def fn(shape, fill_value):
            return func(shape, fill_value, dtype)
        query = f'{str(fn(5, 0))} limit 1;'
    else:

        @heavydb('T[](int32)', T=[dtype], devices=['cpu'])
        def fn(sz):
            z = func(sz, dtype)
            return z
        query = 'select fn(5) limit 1;'

    _, result = heavydb.sql_execute(query)
    # if the execution succeed, it means the return array has type specified by dtype
    assert len(list(result)[0][0]) > 0


dtypes = ('int8_t', 'int16_t', 'int32_t', 'int64_t', 'float', 'double')


@pytest.mark.parametrize('dtype', dtypes)
@pytest.mark.parametrize('func_name', ('full', 'full_like'))
def test_detect_dtype_from_fill_value(heavydb, dtype, func_name):
    heavydb.reset()

    if func_name == 'full':

        @heavydb('T[](int32, T)', T=[dtype], devices=['cpu'])
        def fn(shape, fill_value):
            return array_api.full(shape, fill_value)

        query = str(fn(5, 3))
    else:  # full_like

        @heavydb('T[](T[], T)', T=[dtype], devices=['cpu'])
        def fn(arr, fill_value):
            return array_api.full_like(arr, fill_value)

        datum_type = type_to_type_name(Type.fromstring(dtype))
        query = (f'SELECT fn({dtype_to_col[dtype]}, CAST(3 as {datum_type})) '
                 f'FROM {heavydb.table_name}')

    _, result = heavydb.sql_execute(f'{query} limit 1;')
    # if the execution succeed, it means the return array has type specified by fill_value
    assert len(list(result)[0][0]) > 0




"""Implement Heavydb Column<Array<T>> type support

Heavydb Column<Array<T>> type is the type of input/output column arguments in
UDTFs.
"""

__all__ = ['HeavyDBOutputColumnArrayType', 'HeavyDBColumnArrayType', 'ColumnArray']

import operator
from typing import TypeVar

from llvmlite import ir
from numba.core import cgutils, extending
from numba.core import types as nb_types

from rbc import typesystem

from . import array
from .column import HeavyDBColumnType
from .metatype import HeavyDBMetaType

T = TypeVar('T')


i1 = ir.IntType(1)
i8 = ir.IntType(8)
i8p = i8.as_pointer()
i8pp = i8p.as_pointer()
i64 = ir.IntType(64)
i64p = i64.as_pointer()
void = ir.VoidType()


_COLUMN_PARAM_NAME = 'ColumnArray_inner_type'


class ColumnArray(metaclass=HeavyDBMetaType):
    """
    RBC ``Column<Array<T>>`` type that corresponds to HeavyDB COLUMN<ARRAY<T>>

    In HeavyDB, a Column of type ``Array<T>`` is represented as follows:

    .. code-block:: c

        {
            int8_t* flatbuffer;
            int64_t size;
        }

    """
    def get_item(self, index: int) -> int:
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def set_item(self, index: int, arr: 'array.Array[T]') -> int:
        """
        ``self[index] = arr``

        .. note::
            Only available on ``CPU``
        """

    def concat_item(self, index: int, arr: 'array.Array[T]') -> int:
        """
        Concat array ``arr`` at ``index``.

        .. note::
            Only available on ``CPU``
        """

    def is_null(self, index: int) -> int:
        """
        Check if the array at ``index`` is null.

        .. note::
            Only available on ``CPU``
        """

    def set_null(self, index: int) -> int:
        """
        Set the array at ``index`` to null.

        .. note::
            Only available on ``CPU``
        """

    def __len__(self) -> int:
        """
        Return the length of the array.

        ..  note::
             Only available on ``CPU``
        """


class HeavyDBColumnArrayType(HeavyDBColumnType):

    def postprocess_type(self):
        return self

    @property
    def numba_type(self):
        return ColumnArrayType

    @property
    def numba_pointer_type(self):
        return ColumnArrayPointer

    @property
    def element_type(self):
        return typesystem.Type.fromstring('int8_t')

    @property
    def custom_params(self):
        return {
            **super().custom_params,
            'name': f'Column<Array<{self[0][0][0][0].tostring()}>>',
            _COLUMN_PARAM_NAME: self[0][0],
        }


class HeavyDBOutputColumnArrayType(HeavyDBColumnArrayType):
    """Heavydb OutputColumn type for RBC typesystem.

    OutputColumn is the same as Column<Array<T>> but introduced to distinguish
    the input and output arguments of UDTFs.
    """


class ColumnArrayType(nb_types.Type):
    """Numba type class for HeavyDB buffer structures.
    """

    def __init__(self, dtype):
        array_T = self.__typesystem_type__._params.get(_COLUMN_PARAM_NAME)
        eltype = array_T.tonumba().eltype
        self.dtype = array_T.tonumba()  # struct dtype
        name = f"Column<Array<{eltype}>>"
        super().__init__(name)

    @property
    def key(self):
        return self.dtype

    @property
    def eltype(self):
        """
        Return buffer element dtype.
        """
        return self.dtype.eltype


class ColumnArrayPointer(nb_types.Type):
    """Numba type class for pointers to HeavyDB buffer structures.

    We are not deriving from CPointer because ColumnArrayPointer getitem is
    used to access the data stored in Buffer ptr member.
    """
    mutable = True
    return_as_first_argument = True

    def __init__(self, dtype):
        self.dtype = dtype    # struct dtype
        self.eltype = dtype.eltype  # buffer element dtype
        name = f"Column<Array<{self.eltype}>>*"
        super().__init__(name)

    @property
    def key(self):
        return self.dtype


@extending.intrinsic
def heavydb_column_array_getitem_(typingctx, x, i):
    T = x.eltype
    array = typesystem.Type.fromstring(f'Array<{T}>').tonumba().dtype
    sig = array(x, i)

    def codegen(context, builder, sig, args):
        col, index = args
        fa = cgutils.create_struct_proxy(sig.return_type)(context, builder)
        fnty = ir.FunctionType(void, [i8p, i64, i64, i8pp, i64p, i8p, i64])
        getItem = cgutils.get_or_insert_function(builder.module, fnty,
                                                 "ColumnArray_getItem")
        flatbuffer = builder.extract_value(col, [0])
        ptr = builder.alloca(fa.ptr.type)
        size = builder.alloca(fa.sz.type)
        is_null = builder.alloca(fa.is_null.type)
        expected_numel = i64(-1)
        element_size = i64(array.eltype.bitwidth // 8)
        builder.call(getItem, [flatbuffer, index, expected_numel,
                               builder.bitcast(ptr, i8pp), size,
                               is_null, element_size])
        fa.ptr = builder.load(ptr)
        fa.sz = builder.sdiv(builder.load(size), element_size)
        fa.is_null = builder.load(is_null)
        return fa._getvalue()

    return sig, codegen


@extending.intrinsic
def heavydb_column_array_setitem_(typingctx, x, i, arr):
    sig = nb_types.void(x, i, arr)

    def codegen(context, builder, sig, args):
        col, index, arr_ = args
        fnty = ir.FunctionType(void, [i8p, i64, i8p, i64, i8, i64])
        getItem = cgutils.get_or_insert_function(builder.module, fnty,
                                                 "ColumnArray_setItem")
        flatbuffer = builder.extract_value(col, [0])
        element_size = i64(arr.eltype.bitwidth // 8)
        fa = cgutils.create_struct_proxy(arr)(context, builder, value=arr_)
        builder.call(getItem, [flatbuffer, index, builder.bitcast(fa.ptr, i8p),
                               fa.sz, fa.is_null, element_size])

    return sig, codegen


@extending.intrinsic
def heavydb_column_array_concatitem_(typingctx, x, i, arr):
    sig = nb_types.void(x, i, arr)

    def codegen(context, builder, sig, args):
        col, index, arr_ = args
        fnty = ir.FunctionType(void, [i8p, i64, i8p, i64, i8, i64])
        getItem = cgutils.get_or_insert_function(builder.module, fnty,
                                                 "ColumnArray_concatItem")
        flatbuffer = builder.extract_value(col, [0])
        element_size = i64(arr.eltype.bitwidth // 8)
        fa = cgutils.create_struct_proxy(arr)(context, builder, value=arr_)
        builder.call(getItem, [flatbuffer, index, builder.bitcast(fa.ptr, i8p),
                               fa.sz, fa.is_null, element_size])

    return sig, codegen


@extending.intrinsic
def heavydb_column_array_is_null_(typingctx, x, i):
    sig = nb_types.boolean(x, i)

    def codegen(context, builder, sig, args):
        col, index = args
        fnty = ir.FunctionType(i1, [i8p, i64])
        isNull = cgutils.get_or_insert_function(builder.module, fnty,
                                                "ColumnArray_isNull")
        flatbuffer = builder.extract_value(col, [0])
        return builder.call(isNull, [flatbuffer, index])

    return sig, codegen


@extending.intrinsic
def heavydb_column_array_set_null_(typingctx, x, i):
    sig = nb_types.void(x, i)

    def codegen(context, builder, sig, args):
        col, index = args
        fnty = ir.FunctionType(i1, [i8p, i64])
        setNull = cgutils.get_or_insert_function(builder.module, fnty,
                                                 "ColumnArray_setNull")
        flatbuffer = builder.extract_value(col, [0])
        builder.call(setNull, [flatbuffer, index])

    return sig, codegen


@extending.intrinsic
def deref(typingctx, x):
    sig = x.dtype(x)

    def codegen(context, builder, sig, args):
        [ptr] = args
        return builder.load(ptr)

    return sig, codegen


@extending.intrinsic
def heavydb_column_array_len_(typingctx, x):
    sig = nb_types.int64(x)

    def codegen(context, builder, sig, args):
        [col] = args
        return builder.extract_value(col, [1])

    return sig, codegen


@extending.overload(operator.getitem)
@extending.overload_method(ColumnArrayPointer, 'get_item')
def heavydb_column_array_getitem(x, i):
    if isinstance(x, ColumnArrayPointer):
        def impl(x, i):
            return heavydb_column_array_getitem_(deref(x), i)
        return impl
    elif isinstance(x, ColumnArrayType):
        def impl(x, i):
            return heavydb_column_array_getitem_(x, i)
        return impl


@extending.overload_method(ColumnArrayPointer, 'is_null')
@extending.overload_method(ColumnArrayType, 'is_null')
def heavydb_column_array_ptr_is_null(x, i):
    def impl(x, i):
        return heavydb_column_array_is_null_(deref(x), i)
    return impl


@extending.overload_method(ColumnArrayPointer, 'set_null')
@extending.overload_method(ColumnArrayType, 'set_null')
def heavydb_column_array_ptr_set_null(x, i):
    def impl(x, i):
        return heavydb_column_array_set_null_(deref(x), i)
    return impl


@extending.overload(operator.setitem)
@extending.overload_method(ColumnArrayPointer, 'set_item')
def heavydb_column_array_set_item(x, i, arr):
    if isinstance(x, ColumnArrayPointer):
        def impl(x, i, arr):
            return heavydb_column_array_setitem_(deref(x), i, arr)
        return impl
    elif isinstance(x, ColumnArrayType):
        def impl(x, i, arr):
            return heavydb_column_array_setitem_(x, i, arr)
        return impl


@extending.overload_method(ColumnArrayPointer, 'concat_item')
def heavydb_column_array_concat_item(x, i, arr):
    if isinstance(x, ColumnArrayPointer):
        def impl(x, i, arr):
            return heavydb_column_array_concatitem_(deref(x), i, arr)
        return impl
    elif isinstance(x, ColumnArrayType):
        def impl(x, i, arr):
            return heavydb_column_array_concatitem_(x, i, arr)
        return impl


@extending.overload(len)
def heavydb_column_array_len(x):
    if isinstance(x, ColumnArrayPointer):
        def impl(x):
            return heavydb_column_array_len_(deref(x))
        return impl
    elif isinstance(x, ColumnArrayType):
        def impl(x):
            return heavydb_column_array_len_(x)
        return impl



from rbc.utils import is_localhost, get_local_ip, triple_matches, check_returns_none


def test_is_localhost():
    assert is_localhost(get_local_ip())


def test_triple_matches():
    assert triple_matches('cuda', 'nvptx64-nvidia-cuda')
    assert triple_matches('nvptx64-nvidia-cuda', 'cuda')
    assert triple_matches('cuda32', 'nvptx-nvidia-cuda')
    assert triple_matches('nvptx-nvidia-cuda', 'cuda32')
    assert triple_matches('x86_64-pc-linux-gnu', 'x86_64-unknown-linux-gnu')


def test_check_returns_none():

    def foo():
        return

    assert check_returns_none(foo)

    def foo():
        return None

    assert check_returns_none(foo)

    def foo():
        pass

    assert check_returns_none(foo)

    def foo():
        if 1:
            return
        else:
            return None

    assert check_returns_none(foo)

    def foo():
        return 1

    assert not check_returns_none(foo)

    def foo():
        if 1:
            return
        else:
            return 1

    assert check_returns_none(foo)

    def foo():
        global a
        if a:
            return
        else:
            return 1

    assert not check_returns_none(foo)

    def foo(a=None):
        return a

    assert not check_returns_none(foo)

    def foo(a):
        if a:
            return a + 1
        else:
            return
    assert not check_returns_none(foo)

    def foo(a):
        pass
    assert check_returns_none(foo)

    def foo():
        a = None
        return a
    assert not check_returns_none(foo)  # false-negative



"""Low-level intrinsics to expose external functions
"""

# This code has heavily inspired in the numba.extending.intrisic code
import warnings
from typing import List, Union
from numba.core import extending, funcdesc, types, typing
from rbc.typesystem import Type
from rbc.targetinfo import TargetInfo
from rbc.errors import UnsupportedError
from rbc.utils import validate_devices


class External:

    @classmethod
    def external(cls, signature, *, devices=['CPU']):
        """Parameters
        ----------
        signature : object (str, ctypes function, python callable, numba function)
            Any object convertible to a Numba function via Type.fromobject(...).tonumba()

        Keyword parameters
        ------------------
        devices : list
          Device names ('CPU' or/and 'GPU') for the given set of
          signatures. Default is ['CPU'].
        """
        with TargetInfo.dummy():
            t = Type.fromobject(signature)
        if not t.is_function:
            raise ValueError("signature must represent a function type")
        name = t.name
        if not name:
            raise ValueError(
                f"external function name not specified for signature {signature}"
            )

        # Default devices contains `CPU` only because in most cases
        # the externals are defined for the CPU target.
        devices = validate_devices(devices)
        # Using annotation devices (specified via `| CPU` or `| GPU`) is deprecated.
        # TODO: turn the warning to an exception for rbc version 0.12 or higher
        annotation_devices = validate_devices([d for d in t.annotation()
                                               if d.upper() in {'CPU', 'GPU'}])
        for d in annotation_devices:
            warnings.warn(
                f'Signature {signature} uses deprecated device annotation (`|{d}`).'
                ' Use `devices` kw argument to specify the supported devices of the'
                f' external function `{name}`')
        devices = annotation_devices or devices

        # The key must contain devices bit to avoid spurious errors
        # when externals with the same name but variable device
        # targets are defined (as exemplified in
        # tests/heavydb/test_array.py:test_ptr):
        key = f'{name}-{"-".join(sorted(devices))}'
        obj = cls(key, name, signature, devices)
        obj.register()
        return obj

    def __init__(
        self,
        key: str,
        name: str,
        signature: List[Union[str, types.FunctionType, Type]],
        devices: List[str]
    ):
        """
        Parameters
        ----------
        key : str
            The key of the external function for typing
        name : str
            The name of the external function
        signatures : Dictionary of devices and function signatures
            A device mapping of a list of function type signature
        """
        self.key = key
        self.name = name
        self.signature = signature
        self.devices = devices

    def __str__(self):
        return f'{self.signature} | {" | ".join(self.devices)}'

    def match_signature(self, atypes):
        # Code here is the same found in remotejit.py::Signature::best_match
        target_info = TargetInfo()
        device = "CPU" if target_info.is_cpu else "GPU"
        if device not in self.devices:
            compile_target = target_info.get_compile_target()
            # Raising UnsupportedError will skip compiling the compile
            # target for the given device. This could also be skipped
            # by specifying the desired device in the devices kw
            # argument of a RemoteJIT decorator.
            if device == 'GPU':
                raise UnsupportedError(
                    f'no {device} specific signatures found for the external function'
                    f' `{self.name}` that is referenced by `{compile_target}`.'
                    f' Use devices=["CPU"] when defining `{compile_target.split("__", 1)[0]}`')
            else:
                raise UnsupportedError(
                    f'no {device} specific signatures found for the external function'
                    f' `{self.name}` that is referenced by `{compile_target}`.')

        this_type = Type.fromobject(self.signature)

        if this_type.match(atypes) is not None:
            return this_type

        satypes = ", ".join(map(str, atypes))
        raise TypeError(
            f"{compile_target=}: found no matching function type to the given argument types"
            f" `{satypes}` and device `{device}` while processing `{self.name}`."
            f" Available function type is `{this_type}`."
        )

    def get_codegen(self):
        # lowering
        def codegen(context, builder, sig, args):
            # Need to retrieve the function name again as the function
            # implementation may be device specific
            atypes = tuple(map(Type.fromobject, sig.args))
            t = self.match_signature(atypes)
            fn_name = t.name
            fndesc = funcdesc.ExternalFunctionDescriptor(
                fn_name, sig.return_type, sig.args
            )
            func = context.declare_external_function(builder.module, fndesc)
            return builder.call(func, args)

        return codegen

    def register(self):
        # typing
        def generic(self, args, kws):
            # get the correct signature and function name for the
            # current device
            atypes = tuple(map(Type.fromobject, args))
            t = self.obj.match_signature(atypes)
            TargetInfo().add_external(t.name)

            codegen = self.obj.get_codegen()
            t_numba = t.tonumba()
            extending.lower_builtin(self.key, *t_numba.args)(codegen)
            return t_numba

        ExternalTemplate = type(f"ExternalTemplate_{self.name}",
                                (typing.templates.AbstractTemplate,),
                                dict(generic=generic,
                                     obj=self,
                                     key=self.key))

        typing.templates.infer(ExternalTemplate)
        typing.templates.infer_global(self, types.Function(ExternalTemplate))

    def __call__(self, *args, **kwargs):
        """
        This is only defined to pretend to be a callable from CPython.
        """
        msg = f"{self.name} is not usable in pure-python"
        raise NotImplementedError(msg)


external = External.external



__all__ = ['heavydb_fixture', 'sql_execute']


from abc import abstractproperty
import numpy as np
import os
import pytest
import warnings
import numpy

from packaging.version import Version


def assert_equal(actual, desired):
    """Test equality of actual and desired.

    When both inputs are numpy array or number objects, test equality
    of dtype attributes as well.
    """
    numpy.testing.assert_equal(actual, desired)

    if isinstance(actual, numpy.ndarray) and isinstance(desired, numpy.ndarray):
        numpy.testing.assert_equal(actual.dtype, desired.dtype)
    elif isinstance(actual, numpy.number) and isinstance(desired, numpy.number):
        numpy.testing.assert_equal(actual.dtype, desired.dtype)


def sql_execute(query):
    """Execute a SQL statement to heavydb server using global instance.

    Use when the query does not require registration of new UDF/UDTFs.
    """
    rbc_heavydb = pytest.importorskip('rbc.heavydb')
    heavydb = next(rbc_heavydb.global_heavydb_singleton)
    return heavydb.sql_execute(query)


class _TestTable:

    @classmethod
    def suffix(cls):
        assert cls.__name__.startswith('_')
        assert cls.__name__.endswith('TestTable')
        return cls.__name__[1:-len('TextTable')].lower()

    @abstractproperty
    def sqltypes(self):
        # TODO: for readability, sqltypes should return a mapping from
        # column name -> column type
        pass

    @property
    def colnames(self):
        return self.values.keys()

    @abstractproperty
    def values(self):
        pass

    @property
    def table_defn(self):
        return ',\n'.join(f'{n} {t}' for n, t in zip(self.colnames, self.sqltypes))

    @property
    def require_version(self):
        return (5, 7, 0)


class _DefaultTestTable(_TestTable):

    @classmethod
    def suffix(cls):
        return ''

    @property
    def sqltypes(self):
        return ('FLOAT', 'DOUBLE', 'TINYINT', 'SMALLINT', 'INT', 'BIGINT',
                'BOOLEAN')

    @property
    def values(self):
        return {
            'f4': [0, 1, 2, 3, 4],
            'f8': [0, 1, 2, 3, 4],
            'i1': [0, 1, 2, 3, 4],
            'i2': [0, 1, 2, 3, 4],
            'i4': [0, 1, 2, 3, 4],
            'i8': [0, 1, 2, 3, 4],
            'b': [True, False, True, False, True],
        }


class _10TestTable(_DefaultTestTable):

    @classmethod
    def suffix(cls):
        return '10'

    @property
    def values(self):
        return {
            'f4': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            'f8': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            'i1': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            'i2': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            'i4': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            'i8': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            'b': [True, False, True, False, True, False, True, False, True, False],
        }


class _nullTestTable(_DefaultTestTable):

    @classmethod
    def suffix(cls):
        return "null"

    @property
    def values(self):
        return {
            'f4': [None, 1, 2, None, 4],
            'f8': [0, 1, None, 3, 4],
            'i1': [0, None, 2, 3, None],
            'i2': [None, 1, 2, None, 4],
            'i4': [0, 1, None, 3, 4],
            'i8': [0, None, 2, 3, None],
            'b': [None, False, True, None, True],
        }


class _arrayTestTable(_TestTable):

    @classmethod
    def suffix(cls):
        return "array"

    @property
    def sqltypes(self):
        return ('FLOAT[]', 'DOUBLE[]', 'TINYINT[]', 'SMALLINT[]', 'INT[]', 'BIGINT[]',
                'BOOLEAN[]')

    @property
    def values(self):
        return {
            'f4': [[], [1], [2, 3], [3, 4, 5], [4, 5, 6, 7]],
            'f8': [[], [1], [2, 3], [3, 4, 5], [4, 5, 6, 7]],
            'i1': [[], [1], [2, 3], [3, 4, 5], [4, 5, 6, 7]],
            'i2': [[], [1], [2, 3], [3, 4, 5], [4, 5, 6, 7]],
            'i4': [[], [1], [2, 3], [3, 4, 5], [4, 5, 6, 7]],
            'i8': [[], [1], [2, 3], [3, 4, 5], [4, 5, 6, 7]],
            'b': [[], [False], [True, False], [False, True, False], [True, False, True, False]],
        }


class _arraynullTestTable(_arrayTestTable):

    @classmethod
    def suffix(cls):
        return "arraynull"

    @property
    def values(self):
        return {
            'f4': [None, [1], None, [None, 4, 5], None],
            'f8': [[], None, [None, 3], None, [4, None, 6, 7]],
            'i1': [None, [None], None, [3, None, 5], None],
            'i2': [[], None, [2, None], None, [4, 5, None, 7]],
            'i4': [None, [1], None, [3, 4, None], None],
            'i8': [[], None, [2, 3], None, [None, 5, 6, None]],
            'b': [None, [False], None, [None, True, False], None],
        }


class _TimestampTestTable(_TestTable):

    @classmethod
    def suffix(cls):
        return 'timestamp'

    @property
    def sqltypes(self):
        return ('TIMESTAMP(9)', 'TIMESTAMP(9)', 'TIMESTAMP(9)', 'BIGINT', 'TIMESTAMP(6)')

    @property
    def values(self):
        return {
            "t9": [np.datetime64("1971-01-01 01:01:01.001001001").astype('long'),
                   np.datetime64("1972-02-02 02:02:02.002002002").astype('long'),
                   np.datetime64("1973-03-03 03:03:03.003003003").astype('long')],
            "t9_2": [np.datetime64("2021-01-01 01:01:01.001001001").astype('long'),
                     np.datetime64("2022-02-02 02:02:02.002002002").astype('long'),
                     np.datetime64("2023-03-03 03:03:03.003003003").astype('long')],
            "t9_null": [np.datetime64("1972-02-02 02:02:02.002002002").astype('long'),
                        np.datetime64('NaT').astype('long'),
                        np.datetime64("2037-02-02 02:02:02.002002002").astype('long')],
            "i8_2": [1609462861001001001, 1643767322002002002, 1677812583003003003],
            't6': [np.datetime64("1971-01-01 01:01:01.001001").astype('long'),
                   np.datetime64("1972-02-02 02:02:02.002002").astype('long'),
                   np.datetime64("1973-03-03 03:03:03.003003").astype('long')]
        }


class _TextTestTable(_TestTable):

    @classmethod
    def suffix(cls):
        return 'text'

    @property
    def sqltypes(self):
        return ('TEXT ENCODING DICT(32)', 'TEXT ENCODING DICT(16)',
                'TEXT ENCODING DICT(8)', 'TEXT[] ENCODING DICT(32)',
                'TEXT ENCODING NONE', 'TEXT ENCODING NONE')

    @property
    def values(self):
        return {
            't4': ['foofoo', 'bar', 'fun', 'bar', 'foo'],
            't2': ['foofoo', 'bar', 'fun', 'bar', 'foo'],
            't1': ['fun', 'bar', 'foo', 'barr', 'foooo'],
            's': [['foo', 'bar'], ['fun', 'bar'], ['foo'], ['foo', 'bar'], ['fun', 'bar']],
            'n': ['fun', 'bar', 'foo', 'barr', 'foooo'],
            'n2': ['1', '12', '123', '1234', '12345'],
        }


class _PointTestTable(_TestTable):

    @property
    def sqltypes(self):
        return ("POINT",
                "GEOMETRY(POINT, 4326)",
                "GEOMETRY(POINT, 4326) ENCODING NONE",
                "GEOMETRY(POINT, 900913)")

    @property
    def values(self):
        return {
            'p1': ['POINT(1 2)', 'POINT(9 8)', None],
            'p2': ['POINT(3 4)', 'POINT(7 6)', None],
            'p3': ['POINT(5 6)', 'POINT(5 4)', None],
            'p4': ['POINT(7 8)', 'POINT(3 2)', None],
        }


class _LineStringTestTable(_TestTable):

    @property
    def sqltypes(self):
        return ("LINESTRING",
                "GEOMETRY(LINESTRING, 4326) ENCODING NONE",
                "GEOMETRY(LINESTRING, 4326) ENCODING NONE",
                "GEOMETRY(LINESTRING, 900913)")

    @property
    def values(self):
        return {
            'l1': ['LINESTRING(1 2, 3 5)', 'LINESTRING(9 8, 11 11)', None],
            'l2': ['LINESTRING(3 4, 5 7)', 'LINESTRING(7 6, 9 9)', None],
            'l3': ['LINESTRING(5 6, 7 9)', 'LINESTRING(5 4, 7 7)', None],
            'l4': ['LINESTRING(7 8, 9 11)', 'LINESTRING(3 2, 5 5)', None],
        }


class _MultiLineStringTestTable(_TestTable):

    @property
    def require_version(self):
        return (6, 2, 0)

    @property
    def sqltypes(self):
        return ("MULTILINESTRING",
                "GEOMETRY(MULTILINESTRING, 4326) ENCODING NONE",
                "GEOMETRY(MULTILINESTRING, 4326) ENCODING NONE",
                "GEOMETRY(MULTILINESTRING, 900913)")

    @property
    def values(self):
        return {
            'ml1': ['MULTILINESTRING((1 2,3 4,5 6,7 8,9 10),(2 3,3 4,1 2))',
                    'MULTILINESTRING((0 0,5 0,5 5,0 5))',
                    'MULTILINESTRING((1 2,3 4,5 6,7 8,9 10),(3 4,1 2,2 3),(5 6,7 8,9 10))',
                    None],
            'ml2': ['MULTILINESTRING((0 0,5 0,5 5,0 5),(2 2, 2 1,1 1,1 2))',
                    'MULTILINESTRING((0 0,6 0,6 6,0 6))',
                    'MULTILINESTRING((0 0,5 0,5 5,0 5),(2 2,2 1,1 1,1 2),(0 0,0 1,1 0))',
                    None],
            'ml3': ['MULTILINESTRING((0 0,6 0,6 6,0 6),(3 3,3 2,2 2,2 3))',
                    'MULTILINESTRING((0 0,7 0,7 7,0 7))',
                    'MULTILINESTRING((0 0,6 0,6 6,0 6),(3 3,3 2,2 2,2 3),(0 0,0 1,1 0))',
                    None],
            'ml4': ['MULTILINESTRING((0 0,7 0,7 7,0 7),(4 4,2 4, 2 3,4 2))',
                    'MULTILINESTRING((0 0,4 0,4 4,0 4))',
                    'MULTILINESTRING((0 0,7 0,7 7,0 7),(4 4,2 4, 2 3,4 2),(0 0,0 1,1 0))',
                    None],
        }


class _MultiPointTestTable(_TestTable):

    @property
    def require_version(self):
        return (6, 2, 0)

    @property
    def sqltypes(self):
        return ("MULTIPOINT",
                "GEOMETRY(MULTIPOINT, 4326)",  # uses geoint compression
                "GEOMETRY(MULTIPOINT, 4326) ENCODING NONE",
                "GEOMETRY(MULTIPOINT, 900913)")

    @property
    def values(self):
        return {
            'mp1': ['MULTIPOINT(1 2, 3 5)', 'MULTIPOINT(9 8, 11 11)', None],
            'mp2': ['MULTIPOINT(3 4, 5 7)', 'MULTIPOINT(7 6, 9 9)', None],
            'mp3': ['MULTIPOINT(5 6, 7 9)', 'MULTIPOINT(5 4, 7 7)', None],
            'mp4': ['MULTIPOINT(7 8, 9 11)', 'MULTIPOINT(3 2, 5 5)', None],
        }


class _PolygonTestTable(_TestTable):

    @property
    def sqltypes(self):
        return ("POLYGON",
                "GEOMETRY(POLYGON, 4326)",  # uses geoint compression
                "GEOMETRY(POLYGON, 4326) ENCODING NONE",
                "GEOMETRY(POLYGON, 900913)")

    @property
    def values(self):
        return {
            'p1': ['POLYGON((1 2,3 4,5 6,7 8,9 10),(2 3,3 4,1 2))',
                   'POLYGON((0 0,5 0,5 5,0 5,0 0))',
                   'POLYGON((1 2,3 4,5 6,7 8,9 10),(3 4,1 2,2 3),(5 6,7 8,9 10))',
                   None],
            'p2': ['POLYGON((0 0,5 0,5 5,0 5,0 0),(2 2, 2 1,1 1,1 2,2 2))',
                   'POLYGON((0 0,6 0,6 6,0 6,0 0))',
                   'POLYGON((0 0,5 0,5 5,0 5,0 0),(2 2,2 1,1 1,1 2,2 2),(0 0,0 1,1 0))',
                   None],
            'p3': ['POLYGON((0 0,7 0,7 7,0 7,0 0),(4 4,2 4, 2 3,4 2,4 4))',
                   'POLYGON((0 0,7 0,7 7,0 7,0 0))',
                   'POLYGON((0 0,6 0,6 6,0 6,0 0),(3 3,3 2,2 2,2 3,3 3),(0 0,0 1,1 0))',
                   None],
            'p4': ['POLYGON((0 0,6 0,6 6,0 6,0 0),(3 3,3 2,2 2,2 3,3 3))',
                   'POLYGON((0 0,4 0,4 4,0 4,0 0))',
                   'POLYGON((0 0,7 0,7 7,0 7,0 0),(4 4,2 4, 2 3,4 2,4 4),(0 0,0 1,1 0))',
                   None],
        }


class _MultiPolygonTestTable(_TestTable):

    @property
    def sqltypes(self):
        return ("MULTIPOLYGON",
                "GEOMETRY(MULTIPOLYGON, 4326)",  # uses geoint compression
                "GEOMETRY(MULTIPOLYGON, 4326) ENCODING NONE",
                "GEOMETRY(MULTIPOLYGON, 900913)")

    @property
    def values(self):
        return {
            'mp1': ['MULTIPOLYGON(((1 2,3 4,5 6,7 8,9 10),(2 3,3 4,1 2)))',
                    'MULTIPOLYGON(((0 0,5 0,5 5,0 5,0 0)))',
                    'MULTIPOLYGON(((1 2,3 4,5 6,7 8,9 10),(3 4,1 2,2 3),(5 6,7 8,9 10)))',
                    None],
            'mp2': ['MULTIPOLYGON(((0 0,5 0,5 5,0 5,0 0),(2 2, 2 1,1 1,1 2,2 2)))',
                    'MULTIPOLYGON(((0 0,6 0,6 6,0 6,0 0)))',
                    'MULTIPOLYGON(((0 0,5 0,5 5,0 5,0 0),(2 2,2 1,1 1,1 2,2 2),(0 0,0 1,1 0)))',
                    None],
            'mp3': ['MULTIPOLYGON(((0 0,6 0,6 6,0 6,0 0),(3 3,3 2,2 2,2 3,3 3)))',
                    'MULTIPOLYGON(((0 0,7 0,7 7,0 7,0 0)))',
                    'MULTIPOLYGON(((0 0,6 0,6 6,0 6,0 0),(3 3,3 2,2 2,2 3,3 3),(0 0,0 1,1 0)))',
                    None],
            'mp4': ['MULTIPOLYGON(((0 0,7 0,7 7,0 7,0 0),(4 4,2 4, 2 3,4 2,4 4)))',
                    'MULTIPOLYGON(((0 0,4 0,4 4,0 4,0 0)))',
                    'MULTIPOLYGON(((0 0,7 0,7 7,0 7,0 0),(4 4,2 4, 2 3,4 2,4 4),(0 0,0 1,1 0)))',
                    None],
        }


test_classes = (_DefaultTestTable, _10TestTable, _nullTestTable, _arrayTestTable,
                _arraynullTestTable, _TextTestTable, _TimestampTestTable,
                _PointTestTable, _MultiPointTestTable,
                _LineStringTestTable, _MultiLineStringTestTable,
                _PolygonTestTable, _MultiPolygonTestTable)


test_suffices = [t.suffix() for t in test_classes]


def heavydb_fixture(caller_globals, minimal_version=(0, 0),
                    suffices=test_suffices,
                    load_test_data=True, debug=False):
    """Usage from a rbc/tests/test_xyz.py file:

    .. code-block:: python

       import pytest
       from rbc.tests import heavydb_fixture

       @pytest.fixture(scope='module')
       def heavydb():
           from o in heavydb_fixture(globals()):
               # do some customization here
               yield o

    This fixture creates the following tables:

    f'{heavydb.table_name}' - contains columns f8, f4, i8, i4, i2, i1,
                              b with row size 5.

    f'{heavydb.table_name}10' - contains columns f8, f4, i8, i4, i2,
                                i1, b with row size 10.

    f'{heavydb.table_name}null' - contains columns f8, f4, i8, i4, i2,
                                  i1, b with row size 5, contains null
                                  values

    f'{heavydb.table_name}array' - contains arrays f8, f4, i8, i4, i2,
                                   i1, b with row size 5

    f'{heavydb.table_name}arraynull' - contains arrays f8, f4, i8, i4, i2,
                                       i1, b with row size 5, contains null
                                       values.

    f'{heavydb.table_name}text' - contains text t4, t2, t1, s, n
                                  where 't' prefix is for text encoding dict
                                  and 'n' is for text encoding none.

    f'{heavydb.table_name}timestamp' - contains timestamp t9, t6
                                  where 't' prefix is for timestamp.

    f'{heavydb.table_name}point' - contains columns with GeoPoint

    f'{heavydb.table_name}multipoint' - contains columns with GeoMultiPoint

    f'{heavydb.table_name}linestring' - contains columns with LineString

    f'{heavydb.table_name}multilinestring' - contains columns with MultiLineString

    f'{heavydb.table_name}polygon' - contains columns with Polygon

    f'{heavydb.table_name}multipolygon' - contains columns with MultiPolygon
    """
    rbc_heavydb = pytest.importorskip('rbc.heavydb')
    available_version, reason = rbc_heavydb.is_available()

    def require_version(version, message=None, label=None):
        """Execute pytest.skip(...) if version is older than available_version.

        Some tests can be run only when using heavydb server built
        from a particular branch of heavydb.  So, when the specified
        version and the heavydb version match exactly and these
        correspond to the current development version, if the
        specified label does not match with the value of envrinment
        variable HEAVYDB_DEV_LABEL, then the corresponing test will
        be skipped. Use label 'docker-dev' when using heavydb dev
        docker image.

        """
        # The available version (of the heavydb server) has date and
        # hash bits, however, these are useless for determining the
        # version ordering (in the case of available_version[:3] ==
        # version) because the date corresponds to the date of
        # building the server and the hash corresponds to some commit
        # of some repository (heavydb or heavydb-internal) and it
        # does not provide easy date information.
        #
        # The condition available_version[:3] == version can appear in
        # the following cases (given in the order of from newer to
        # older):
        # 1. heavydb is built against a heavydb-internal PR branch
        # (assuming it is rebased against master)
        # 2. heavydb is built against heavydb-internal master branch
        # 3. heavydb is built against heavydb master branch
        # 4. heavydb is built against heavydb dev docker
        # 5. heavydb is built against heavydb/heavydb-internal release tag
        #
        # rbc testing suite may use features that exists in the head
        # of the above list but not in the tail of it. So we have a
        # problem of deciding if a particular test should be disabled
        # or not for a given case while there is no reliable way to
        # tell from heavydb version if the server has the particular
        # feature or not. To resolve this, we use label concept as
        # explained in the doc-string.
        #

        if not available_version:
            pytest.skip(reason)
        # Requires update when heavydb-internal bumps up version number:
        current_development_version = Version("6.4.0")

        curr_version = Version('.'.join(map(str, available_version[:2])))

        if curr_version > current_development_version:
            warnings.warn(f'{curr_version}) is newer than development version'
                          f' ({current_development_version}), please update the latter!')

        assert isinstance(version, tuple)
        version = Version('.'.join(map(str, version)))
        if version > curr_version:
            _reason = f'test requires version {version} or newer, got {available_version}'
            if message is not None:
                _reason += f': {message}'
            pytest.skip(_reason)

        if label is not None:
            env_label = os.environ.get('HEAVYDB_DEV_LABEL')
            if env_label and label == 'docker-dev':
                # docker-dev is some older master, so it must work
                # with the current master as well as with branches based on the current master.
                label = 'master'
            if label == 'master' and env_label and env_label != 'docker-dev':
                # assuming that the branch given in the label is
                # up-to-date with respect to master branch. If it is
                # not, one should rebase the branch against the
                # master.
                label = env_label
            if env_label is None:
                warnings.warn('Environment does not specify label (HEAVYDB_DEV_LABEL is unset).'
                              ' Tests with development labels will not be run.')
            if env_label != label:
                _reason = (f'test requires version {version} with label {label},'
                           f' got {curr_version} with label {env_label}')
                if message is not None:
                    _reason += f': {message}'
                pytest.skip(_reason)

            if version < curr_version:
                # in the case the branch given in the label was never
                # merged, consider removing the corresponding test
                warnings.warn(f'detected test requiring {version} with out-of-date label {label}.'
                              ' Please reset test label to None.')

    require_version(minimal_version)

    filename = caller_globals['__file__']
    table_name = os.path.splitext(os.path.basename(filename))[0]

    config = rbc_heavydb.get_client_config(debug=debug)
    m = rbc_heavydb.RemoteHeavyDB(**config)
    m.require_version = require_version

    if not load_test_data:
        yield m
        return

    # todo: TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://docs.heavy.ai/sql/data-definition-ddl/datatypes-and-fixed-encoding
    curr_version = Version('.'.join(map(str, m.version[:2])))
    for cls in test_classes:
        suffix = cls.suffix()
        if suffix in suffices:
            obj = cls()
            version = Version('.'.join(map(str, obj.require_version)))
            if version > curr_version:
                continue
            m.sql_execute(f'DROP TABLE IF EXISTS {table_name}{suffix}')
            m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name}{suffix} ({obj.table_defn})')
            m.load_table_columnar(f'{table_name}{suffix}', **obj.values)

    m.table_name = table_name
    yield m
    for suffix in suffices:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}{suffix}')



"""RBC GeoPoint type that corresponds to HeavyDB type GEOPOINT.
"""

__all__ = ["HeavyDBGeoPointType", "GeoPoint"]

from .geo_nested_array import (GeoNestedArray, GeoNestedArrayNumbaType,
                               HeavyDBGeoNestedArray)


class GeoPointNumbaType(GeoNestedArrayNumbaType):
    def __init__(self):
        super().__init__(name="GeoPointNumbaType")


class HeavyDBGeoPointType(HeavyDBGeoNestedArray):
    """Typesystem type class for HeavyDB buffer structures."""

    @property
    def type_name(self):
        return "GeoPoint"


class GeoPoint(GeoNestedArray):
    """
    RBC ``GeoPoint`` type that corresponds to HeavyDB type ``POINT``.

    .. code-block:: c

        struct Point2D {
            double x;
            double y;
        }
    """

    @property
    def x(self) -> float:
        """
        Return ``x`` coordinate
        """

    @property
    def y(self) -> float:
        """
        Return ``y`` coordinate
        """



import os
import pytest
from rbc.tests import heavydb_fixture
from rbc.external import external


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(5, 6), debug=not True, suffices=['']):
        yield o


@pytest.fixture(scope='module')
def boston_house_prices(heavydb):
    # Upload Boston house prices to server, notice that we collect all
    # row values expect the last one (MEDV) to a FLOAT array.

    csv_file = os.path.join(os.path.dirname(__file__), '..', 'boston_house_prices.csv')
    data0 = []
    medv0 = []
    for i, line in enumerate(open(csv_file).readlines()):
        line = line.strip().replace(' ', '')
        if i == 0:
            header = line.split(',')
        else:
            row = list(map(float, line.split(',')))
            assert len(row) == len(header)
            data0.append(row[:-1])
            medv0.append(row[-1])
    table_name = f'{heavydb.table_name}bhp'

    heavydb.sql_execute(f'DROP TABLE IF EXISTS {table_name}')
    heavydb.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name} (data FLOAT[], medv FLOAT);')
    heavydb.load_table_columnar(table_name, data=data0, medv=medv0)

    yield heavydb

    heavydb.sql_execute(f'DROP TABLE IF EXISTS {table_name}')


def test_boston_house_prices(heavydb, boston_house_prices):
    if heavydb.compiler is None:
        pytest.skip('test requires C/C++ to LLVM IR compiler')

    treelite = pytest.importorskip("treelite")
    xgboost = pytest.importorskip("xgboost")

    device = 'cpu'
    import numpy as np
    import tempfile

    # Get training data from server:
    table_name = f'{heavydb.table_name}bhp'

    descr, result = heavydb.sql_execute(
        f'SELECT rowid, data, medv FROM {table_name} ORDER BY rowid LIMIT 50')
    result = list(result)
    medv = np.array([medv for _, data, medv in result])
    data = np.array([data for _, data, medv in result])
    assert len(medv) == 50

    # Train model using xgboost
    dtrain = xgboost.DMatrix(data, label=medv)
    params = {'max_depth': 3,
              'eta': 1,
              'objective': 'reg:squarederror',
              'eval_metric': 'rmse'}
    bst = xgboost.train(params, dtrain, len(medv), [(dtrain, 'train')])

    # Compile model to C
    working_dir = tempfile.mkdtemp()
    model = treelite.Model.from_xgboost(bst)
    model.compile(working_dir)
    model_c = open(os.path.join(working_dir, 'main.c')).read()
    # The C model implements
    #  float predict(union Entry* data, int pred_margin)
    # but we wrap it using
    model_c += '''
#ifdef __cplusplus
extern "C" {
#endif
    float predict_float(float* data, int pred_margin) {
      return predict((union Entry*)data, pred_margin);
    }
#ifdef __cplusplus
}
#endif
    '''
    predict_float = external('float predict_float(float*, int32)')
    # to make UDF construction easier. Notice that predict_float can
    # be now called from a UDF.

    # Define predict function as UDF. Notice that the xgboost null
    # values are different from heavydb null values, so we'll remap
    # before calling predict_float:
    null_value = np.int32(-1).view(np.float32)

    @heavydb('float(float[], int32)', devices=[device])
    def predict(data, pred_margin):
        for i in range(len(data)):
            if data.is_null(i):
                data[i] = null_value
        return predict_float(data.ptr(), pred_margin)

    # Compile C model to LLVM IR. In future, we might want this
    # compilation to happen in the server side as the client might not
    # have clang compiler installed.
    model_llvmir = heavydb.compiler(model_c, flags=['-I' + working_dir])

    # RBC will link_in the LLVM IR module
    heavydb.user_defined_llvm_ir[device] = model_llvmir

    # Call predict on data in the server:
    descr, result = heavydb.sql_execute('SELECT rowid, predict(data, 2) FROM'
                                        f' {table_name} ORDER BY rowid')
    result = list(result)
    predict_medv = np.array([r[1] for r in result])

    # predict on the first 50 elements should be close to training labels
    error = abs(predict_medv[:len(medv)] - medv).max()/len(medv)
    assert error < 1e-4, error



from llvmlite import ir
from numba import types as nb_types
from numba.core import cgutils, extending

from rbc.errors import RequireLiteralValue

i8 = ir.IntType(8)
i8p = i8.as_pointer()


@extending.intrinsic
def deref(typingctx, x):
    sig = x.dtype(x)

    def codegen(context, builder, sig, args):
        [ptr] = args
        return builder.load(ptr)

    return sig, codegen


@extending.intrinsic
def ref(typingctx, obj):
    sig = nb_types.CPointer(obj)(obj)

    def codegen(context, builder, sig, args):
        [obj] = args
        fa = context.make_helper(builder, sig.args[0], value=obj)
        return fa._getpointer()

    return sig, codegen


@extending.intrinsic
def as_voidptr(typingctx, x):
    ptr = nb_types.CPointer(nb_types.int8)
    sig = ptr(x)

    def codegen(context, builder, sig, args):
        [ptr] = args
        return builder.bitcast(ptr, i8p)

    return sig, codegen


@extending.intrinsic
def get_alloca(typingctx, x):
    ptr = nb_types.CPointer(x)
    sig = ptr(x)

    def codegen(context, builder, sig, args):
        [val] = args
        fa = context.make_helper(builder, sig.args[0], value=val)
        return builder.bitcast(fa._getpointer(), i8p)

    return sig, codegen


@extending.intrinsic
def global_str_constant(typingctx, identifier, msg):
    ptr = nb_types.CPointer(nb_types.int8)
    sig = ptr(identifier, msg)

    if not isinstance(msg, nb_types.StringLiteral):
        raise RequireLiteralValue(f"expected StringLiteral but got {type(msg).__name__}")

    def codegen(context, builder, sig, args):
        msg_bytes = msg.literal_value.encode('utf-8')
        msg_const = cgutils.make_bytearray(msg_bytes + b'\0')
        msg_global_var = cgutils.global_constant(
            builder.module,
            identifier.literal_value,
            msg_const)
        msg_ptr = builder.bitcast(msg_global_var, i8p)
        return msg_ptr

    return sig, codegen



# Author: Pearu Peterson
# Created: February 2019

import functools
import os
import re
import warnings
from collections import defaultdict
from contextlib import contextmanager
from typing import Optional

import llvmlite.binding as llvm
from llvmlite import ir
from numba.core import cgutils, codegen, compiler, compiler_lock, cpu
from numba.core import errors as nb_errors
from numba.core import extending, imputils, registry, sigutils, typing

from rbc.externals import stdio
from rbc.nrt import create_nrt_functions

from .errors import UnsupportedError
from .libfuncs import Library
from .targetinfo import TargetInfo

int32_t = ir.IntType(32)
int1_t = ir.IntType(1)


def find_at_word(text: str) -> Optional[str]:
    """
    Find called function name from a CallInst to a bitcast.

    Example usage

        >>> text = "i8* (i64)* bitcast (%Struct.MemInfo.5* (i64)* @NRT_MemInfo_alloc_safe to i8* (i64)*)"  # noqa: E501
        >>> find_at_word(text)
        "NRT_MemInfo_alloc_safe
        >>> text = "%.5 = bitcast i8* %info to void (i8*, i64, i8*)*"
        >>> find_at_word(text)
        None

    """
    pattern = r"@\w+"
    words = re.findall(pattern, text)
    if words == []:
        return None
    assert len(words) == 1
    [word] = words
    return word[1:]


def get_called_functions(library, funcname=None, debug=False):
    result = defaultdict(set)
    module = library._final_module
    if funcname is None:
        for df in library.get_defined_functions():
            for k, v in get_called_functions(library, df.name, debug).items():
                result[k].update(v)
        return result

    func = module.get_function(funcname)
    assert func.name == funcname, (func.name, funcname)
    result['defined'].add(funcname)
    for block in func.blocks:
        for instruction in block.instructions:
            if instruction.opcode == 'call':
                instr = list(instruction.operands)[-1]
                name = instr.name
                try:
                    f = module.get_function(name)
                except NameError:
                    if 'bitcast' not in str(instr):
                        raise  # re-raise

                    # attempt to find caller symbol in instr
                    name = find_at_word(str(instr))
                    if name is None:
                        # ignore call to function pointer
                        msg = f'Ignoring call to bitcast instruction:\n{instr}'
                        if debug:
                            warnings.warn(msg)
                        continue
                    f = module.get_function(name)

                if name.startswith('llvm.'):
                    result['intrinsics'].add(name)
                elif f.is_declaration:
                    found = False
                    for lib in library._linking_libraries:
                        for df in lib.get_defined_functions():
                            if name == df.name:
                                result['defined'].add(name)
                                result['libraries'].add(lib)
                                found = True
                                for k, v in get_called_functions(lib, name, debug).items():
                                    result[k].update(v)
                                break
                        if found:
                            break
                    if not found:
                        result['declarations'].add(name)
                else:
                    result['defined'].add(name)
                    for k, v in get_called_functions(library, name, debug).items():
                        result[k].update(v)
    return result


# ---------------------------------------------------------------------------

class JITRemoteCodeLibrary(codegen.JITCodeLibrary):
    """JITRemoteCodeLibrary was introduce to prevent numba from calling functions
    that checks if the module is final. See xnd-project/rbc issue #87.
    """

    def get_pointer_to_function(self, name):
        """We can return any random number here! This is just to prevent numba from
        trying to check if the symbol given by "name" is defined in the module.
        In cases were RBC is calling an external function (i.e. allocate_varlen_buffer)
        the symbol will not be defined in the module, resulting in an error.
        """
        return 0

    def _finalize_specific(self):
        """Same as codegen.JITCodeLibrary._finalize_specific but without
        calling _ensure_finalize at the end
        """
        self._codegen._scan_and_fix_unresolved_refs(self._final_module)


class JITRemoteCodegen(codegen.JITCPUCodegen):
    _library_class = JITRemoteCodeLibrary

    def _get_host_cpu_name(self):
        target_info = TargetInfo()
        return target_info.device_name

    def _get_host_cpu_features(self):
        target_info = TargetInfo()
        features = target_info.device_features
        server_llvm_version = target_info.llvm_version
        if server_llvm_version is None or target_info.is_gpu:
            return ''
        client_llvm_version = llvm.llvm_version_info

        # See https://github.com/xnd-project/rbc/issues/45
        remove_features = {
            (12, 12): [], (11, 11): [], (10, 10): [], (9, 9): [], (8, 8): [],
            (11, 8): ['tsxldtrk', 'amx-tile', 'amx-bf16', 'serialize', 'amx-int8',
                      'avx512vp2intersect', 'tsxldtrk', 'amx-tile', 'amx-bf16',
                      'serialize', 'amx-int8', 'avx512vp2intersect', 'tsxldtrk',
                      'amx-tile', 'amx-bf16', 'serialize', 'amx-int8',
                      'avx512vp2intersect', 'cx8', 'enqcmd', 'avx512bf16'],
            (11, 10): ['tsxldtrk', 'amx-tile', 'amx-bf16', 'serialize', 'amx-int8'],
            (9, 11): ['sse2', 'cx16', 'sahf', 'tbm', 'avx512ifma', 'sha',
                      'gfni', 'fma4', 'vpclmulqdq', 'prfchw', 'bmi2', 'cldemote',
                      'fsgsbase', 'ptwrite', 'xsavec', 'popcnt', 'mpx',
                      'avx512bitalg', 'movdiri', 'xsaves', 'avx512er',
                      'avx512vnni', 'avx512vpopcntdq', 'pconfig', 'clwb',
                      'avx512f', 'clzero', 'pku', 'mmx', 'lwp', 'rdpid', 'xop',
                      'rdseed', 'waitpkg', 'movdir64b', 'sse4a', 'avx512bw',
                      'clflushopt', 'xsave', 'avx512vbmi2', '64bit', 'avx512vl',
                      'invpcid', 'avx512cd', 'avx', 'vaes', 'cx8', 'fma', 'rtm',
                      'bmi', 'enqcmd', 'rdrnd', 'mwaitx', 'sse4.1', 'sse4.2', 'avx2',
                      'fxsr', 'wbnoinvd', 'sse', 'lzcnt', 'pclmul', 'prefetchwt1',
                      'f16c', 'ssse3', 'sgx', 'shstk', 'cmov', 'avx512vbmi',
                      'avx512bf16', 'movbe', 'xsaveopt', 'avx512dq', 'adx',
                      'avx512pf', 'sse3'],
            (9, 8): ['cx8', 'enqcmd', 'avx512bf16'],
            (14, 11): ['crc32', 'uintr', 'widekl', 'avxvnni', 'avx512fp16', 'kl', 'hreset'],
        }.get((server_llvm_version[0], client_llvm_version[0]), None)
        if remove_features is None:
            warnings.warn(
                f'{type(self).__name__}._get_host_cpu_features: `remove_features` dictionary'
                ' requires an update: detected different LLVM versions in server '
                f'{server_llvm_version} and client {client_llvm_version}.'
                f' CPU features: {features}.')
        else:
            features += ','
            for f in remove_features:
                features = features.replace('+' + f + ',', '').replace('-' + f + ',', '')
            features.rstrip(',')
        return features

    def _customize_tm_options(self, options):
        super()._customize_tm_options(options)
        # fix reloc_model as the base method sets it using local target
        target_info = TargetInfo()
        if target_info.arch.startswith('x86'):
            reloc_model = 'static'
        else:
            reloc_model = 'default'
        options['reloc'] = reloc_model

    def set_env(self, env_name, env):
        return None


class JITRemoteTypingContext(typing.Context):
    def load_additional_registries(self):
        self.install_registry(typing.templates.builtin_registry)
        super().load_additional_registries()


class JITRemoteTargetContext(cpu.CPUContext):

    @compiler_lock.global_compiler_lock
    def init(self):
        target_info = TargetInfo()
        self.address_size = target_info.bits
        self.is32bit = (self.address_size == 32)
        self._internal_codegen = JITRemoteCodegen("numba.exec")

    def load_additional_registries(self):
        self.install_registry(imputils.builtin_registry)
        super().load_additional_registries()

    def get_executable(self, library, fndesc, env):
        return None

    def post_lowering(self, mod, library):
        pass


# ---------------------------------------------------------------------------
# Code generation methods


@contextmanager
def replace_numba_internals_hack():
    # Hackish solution to prevent numba from calling _ensure_finalize. See issue #87
    _internal_codegen_bkp = registry.cpu_target.target_context._internal_codegen
    registry.cpu_target.target_context._internal_codegen = JITRemoteCodegen("numba.exec")
    yield
    registry.cpu_target.target_context._internal_codegen = _internal_codegen_bkp


def make_wrapper(fname, atypes, rtype, cres, target: TargetInfo, verbose=False):
    """Make wrapper function to numba compile result.

    The compilation result contains a function with prototype::

      <status> <function name>(<rtype>** result, <arguments>)

    The make_wrapper adds a wrapper function to compilation result
    library with the following prototype::

      <rtype> <fname>(<arguments>)

    or, if <rtype>.return_as_first_argument == True, then

      void <fname>(<rtype>* <arguments>)

    Parameters
    ----------
    fname : str
      Function name.
    atypes : tuple
      A tuple of argument Numba types.
    rtype : numba.Type
      The return Numba type
    cres : CompileResult
      Numba compilation result.
    verbose: bool
      When True, insert printf statements for debugging errors. For
      CUDA target this feature is disabled.

    """
    fndesc = cres.fndesc
    module = cres.library.create_ir_module(fndesc.unique_name)
    context = cres.target_context
    ll_argtypes = [context.get_value_type(ty) for ty in atypes]
    ll_return_type = context.get_value_type(rtype)

    return_as_first_argument = getattr(rtype, 'return_as_first_argument', False)
    assert isinstance(return_as_first_argument, bool)
    if return_as_first_argument:
        wrapty = ir.FunctionType(ir.VoidType(),
                                 [ll_return_type] + ll_argtypes)
        wrapfn = ir.Function(module, wrapty, fname)
        builder = ir.IRBuilder(wrapfn.append_basic_block('entry'))
        fnty = context.call_conv.get_function_type(rtype, atypes)
        fn = ir.Function(builder.module, fnty, cres.fndesc.llvm_func_name)
        status, out = context.call_conv.call_function(
            builder, fn, rtype, atypes, wrapfn.args[1:])
        with cgutils.if_unlikely(builder, status.is_error):
            if verbose and target.is_cpu:
                cgutils.printf(builder,
                               f"rbc: {fname} failed with status code %i\n",
                               status.code)
                stdio._cg_fflush(builder)
            builder.ret_void()
        # does a "deepcopy"
        rtype.deepcopy(context, builder, out, wrapfn.args[0])
        builder.ret_void()
    else:
        wrapty = ir.FunctionType(ll_return_type, ll_argtypes)
        wrapfn = ir.Function(module, wrapty, fname)
        builder = ir.IRBuilder(wrapfn.append_basic_block('entry'))
        fnty = context.call_conv.get_function_type(rtype, atypes)
        fn = ir.Function(builder.module, fnty, cres.fndesc.llvm_func_name)
        status, out = context.call_conv.call_function(
            builder, fn, rtype, atypes, wrapfn.args)
        if verbose and target.is_cpu:
            with cgutils.if_unlikely(builder, status.is_error):
                cgutils.printf(builder,
                               f"rbc: {fname} failed with status code %i\n",
                               status.code)
                stdio._cg_fflush(builder)
        builder.ret(out)

    cres.library.add_ir_module(module)


def compile_instance(func, sig,
                     target: TargetInfo,
                     typing_context,
                     target_context,
                     pipeline_class,
                     main_library,
                     debug=False):
    """Compile a function with given signature. Return function name when
    succesful.
    """
    flags = compiler.Flags()
    flags.no_compile = True
    flags.no_cpython_wrapper = True
    flags.no_cfunc_wrapper = True
    flags.nrt = True

    fname = func.__name__ + sig.mangling()
    args, return_type = sigutils.normalize_signature(
        sig.tonumba(bool_is_int8=True))
    target.set_compile_target(fname)
    try:
        cres = compiler.compile_extra(typingctx=typing_context,
                                      targetctx=target_context,
                                      func=func,
                                      args=args,
                                      return_type=return_type,
                                      flags=flags,
                                      library=main_library,
                                      locals={},
                                      pipeline_class=pipeline_class)
        target.set_compile_target(None)
    except UnsupportedError as msg:
        target.set_compile_target(None)
        warnings.warn(f'Skipping {fname}: {msg.args[1]}')
        return
    except (nb_errors.TypingError, nb_errors.LoweringError) as msg:
        target.set_compile_target(None)
        for m in re.finditer(r'UnsupportedError(.*?)\n', str(msg), re.S):
            warnings.warn(f'Skipping {fname}:{m.group(0)[18:]}')
            break
        else:
            raise
        return
    except Exception:
        target.set_compile_target(None)
        raise

    result = get_called_functions(cres.library, cres.fndesc.llvm_func_name, debug)

    for f in result['declarations']:
        if target.supports(f):
            continue
        warnings.warn(f'Skipping {fname} that uses undefined function `{f}`')
        return

    nvvmlib = Library.get('nvvm')
    llvmlib = Library.get('llvm')
    for f in result['intrinsics']:
        if target.is_gpu:
            if f in nvvmlib:
                continue

        if target.is_cpu:
            if f in llvmlib:
                continue

        warnings.warn(f'Skipping {fname} that uses unsupported intrinsic `{f}`')
        return

    make_wrapper(fname, args, return_type, cres, target, verbose=debug)

    main_module = main_library._final_module
    for lib in result['libraries']:
        main_module.link_in(
            lib._get_module_for_linking(), preserve=True,
        )

    return fname


def add_metadata_flag(main_library, **kwargs):
    module = ir.Module()
    mflags = module.add_named_metadata('llvm.module.flags')
    override_flag = int32_t(4)
    for flag_name, flag_value in kwargs.items():
        flag = module.add_metadata([override_flag, flag_name, int1_t(flag_value)])
        mflags.add(flag)
    main_library.add_ir_module(module)


def read_unicodetype_db():

    @functools.lru_cache()
    def _read_file():
        unicode_file = os.path.join(os.path.dirname(__file__),
                                    'unicodetype_db.ll')
        with open(unicode_file, 'r') as f:
            s = f.read()
        return s

    return llvm.parse_assembly(_read_file())


def compile_to_LLVM(functions_and_signatures,
                    target_info: TargetInfo,
                    pipeline_class=compiler.Compiler,
                    user_defined_llvm_ir=None,
                    debug=False):
    """Compile functions with given signatures to target specific LLVM IR.

    Parameters
    ----------
    functions_and_signatures : list
      Specify a list of Python function and its signatures pairs.
    target : TargetInfo
      Specify target device information.
    user_defined_llvm_ir : {None, str, ModuleRef}
      Specify user-defined LLVM IR module that is linked in to the
      returned module.
    debug : bool

    Returns
    -------
    module : llvmlite.binding.ModuleRef
      LLVM module instance. To get the IR string, use `str(module)`.

    """
    target_desc = registry.cpu_target

    typing_context = JITRemoteTypingContext()
    target_context = JITRemoteTargetContext(typing_context)

    nrt_module = create_nrt_functions(target_context, debug=debug)
    unicodetype_db = read_unicodetype_db()

    # Bring over Array overloads (a hack):
    target_context._defns = target_desc.target_context._defns

    with replace_numba_internals_hack():
        codegen = target_context.codegen()
        main_library = codegen.create_library('rbc.irtools.compile_to_IR')
        main_module = main_library._final_module

        if user_defined_llvm_ir is not None:
            if isinstance(user_defined_llvm_ir, str):
                user_defined_llvm_ir = llvm.parse_assembly(user_defined_llvm_ir)
            assert isinstance(user_defined_llvm_ir, llvm.ModuleRef)
            main_module.link_in(user_defined_llvm_ir, preserve=True)

        main_module.link_in(unicodetype_db)
        main_library.add_ir_module(nrt_module)

        succesful_fids = []
        function_names = []
        for func, signatures in functions_and_signatures:
            for fid, sig in signatures.items():
                fname = compile_instance(func, sig, target_info, typing_context,
                                         target_context, pipeline_class,
                                         main_library,
                                         debug=debug)
                if fname is not None:
                    succesful_fids.append(fid)
                    function_names.append(fname)

        add_metadata_flag(main_library,
                          pass_column_arguments_by_value=0,
                          manage_memory_buffer=1)
        main_library._optimize_final_module()

        # Remove unused defined functions and declarations
        used_symbols = defaultdict(set)
        for fname in function_names:
            for k, v in get_called_functions(main_library, fname, debug).items():
                used_symbols[k].update(v)

        all_symbols = get_called_functions(main_library, debug=debug)

        unused_symbols = defaultdict(set)
        for k, lst in all_symbols.items():
            if k == 'libraries':
                continue
            for fn in lst:
                if fn not in used_symbols[k]:
                    unused_symbols[k].add(fn)

        changed = False
        for f in main_module.functions:
            fn = f.name
            if fn.startswith('llvm.'):
                if f.name in unused_symbols['intrinsics']:
                    f.linkage = llvm.Linkage.external
                    changed = True
            elif f.is_declaration:
                if f.name in unused_symbols['declarations']:
                    f.linkage = llvm.Linkage.external
                    changed = True
            else:
                if f.name in unused_symbols['defined']:
                    f.linkage = llvm.Linkage.private
                    changed = True

        # TODO: determine unused global_variables and struct_types

        if changed:
            main_library._optimize_final_module()

        main_module.verify()
        main_library._finalized = True
        main_module.triple = target_info.triple
        main_module.data_layout = target_info.datalayout

    return main_module, succesful_fids


def compile_IR(ir):
    """Return execution engine with IR compiled in.

    Parameters
    ----------
    ir : str
      Specify LLVM IR code as a string.

    Returns
    -------
    engine :
      Execution engine.


    Examples
    --------

        To get the address of the compiled functions, use::

          addr = engine.get_function_address("<function name>")
    """
    triple = re.search(
        r'target\s+triple\s*=\s*"(?P<triple>[-\d\w\W_]+)"\s*$',
        ir, re.M).group('triple')

    # Create execution engine
    llvm.initialize()
    llvm.initialize_all_targets()
    llvm.initialize_all_asmprinters()

    target = llvm.Target.from_triple(triple)
    target_machine = target.create_target_machine()
    backing_mod = llvm.parse_assembly("")
    engine = llvm.create_mcjit_compiler(backing_mod, target_machine)

    # Create LLVM module and compile
    mod = llvm.parse_assembly(ir)
    mod.verify()
    engine.add_module(mod)
    engine.finalize_object()
    engine.run_static_constructors()

    return engine


def IS_CPU():
    pass


@extending.overload(IS_CPU, inline="always")
def is_cpu_impl():
    target_info = TargetInfo()
    if target_info.is_cpu:
        return lambda: True
    else:
        return lambda: False


def IS_GPU():
    pass


@extending.overload(IS_GPU, inline="always")
def is_gpu_impl():
    target_info = TargetInfo()
    if target_info.is_gpu:
        return lambda: True
    else:
        return lambda: False




import pytest

cudart = pytest.importorskip('rbc.libcudart_ctypes_SKIP')

driver_version = cudart.get_cuda_versions()[0]
if driver_version == 0:
    pytest.importorskip('rbc.libcudart_ctypes_NODRIVER',
                        reason='CUDA driver not installed')

if cudart.get_device_count() == 0:
    pytest.importorskip('rbc.libcudart_ctypes_NODEVICE',
                        reason='CUDA device missing')


def test_get_device_count():
    assert cudart.get_device_count() > 0


def test_get_cuda_device_properties():
    props = cudart.get_cuda_device_properties(0)
    assert props['major'] >= 2



"""
RBC TextEncodingNone type that corresponds to HeavyDB type TEXT ENCODED NONE.
"""

__all__ = ['TextEncodingNonePointer', 'TextEncodingNone', 'HeavyDBTextEncodingNoneType']

import operator
from typing import Union

from llvmlite import ir
from numba.core import cgutils, extending
from numba.core import types as nb_types
from numba.core.pythonapi import PY_UNICODE_1BYTE_KIND
from numba.cpython.unicode import _empty_string, _set_code_point

from rbc import typesystem

from .allocator import allocate_varlen_buffer
from .array import ArrayPointer
from .buffer import Buffer, HeavyDBBufferType, heavydb_buffer_constructor

i8 = ir.IntType(8)
i8p = i8.as_pointer()
i32 = ir.IntType(32)
i64 = ir.IntType(64)
void = ir.VoidType()
int64_t = ir.IntType(64)


class HeavyDBTextEncodingNoneType(HeavyDBBufferType):
    """HeavyDB TextEncodingNone type for RBC typesystem.
    """

    @property
    def numba_pointer_type(self):
        return TextEncodingNonePointer

    @classmethod
    def preprocess_args(cls, args):
        element_type = typesystem.Type.fromstring('char8')
        return ((element_type,),)

    @property
    def buffer_extra_members(self):
        return ('bool is_null',)

    def match(self, other):
        if type(self) is type(other):
            return self[0] == other[0]
        if other.is_pointer and other[0].is_char and other[0].bits == 8:
            return 1
        if other.is_string:
            return 2


class TextEncodingNonePointer(ArrayPointer):
    def deepcopy(self, context, builder, val, retptr):
        zero, one, two = i32(0), i32(1), i32(2)

        src = builder.load(builder.gep(val, [zero, zero]))
        sz = builder.load(builder.gep(val, [zero, one]))
        is_null = builder.load(builder.gep(val, [zero, two]))

        dst = allocate_varlen_buffer(builder, sz, i64(1))
        cgutils.raw_memcpy(builder, dst=dst, src=src, count=sz, itemsize=i64(1))
        # string is null terminated
        builder.store(i8(0), builder.gep(dst, [sz]))

        builder.store(dst, builder.gep(retptr, [zero, zero]))
        builder.store(sz, builder.gep(retptr, [zero, one]))
        builder.store(is_null, builder.gep(retptr, [zero, two]))


class TextEncodingNone(Buffer):
    """
    RBC ``TextEncodingNone`` type that corresponds to HeavyDB type TEXT ENCODED NONE.
    HeavyDB TextEncodingNone represents the following structure:

    .. code-block:: c

        struct TextEncodingNone {
            char* ptr;
            size_t size;
            int8_t padding;
        }

    .. code-block:: python

        from rbc.heavydb import TextEncodingNone
        @heavydb('TextEncodingNone(int32, int32)')
        def make_abc(first, n):
            r = TextEncodingNone(n)
            for i in range(n):
                r[i] = first + i
            return r

    .. code-block:: python

        from rbc.heavydb import TextEncodingNone
        @heavydb('TextEncodingNone()')
        def make_text():
            return TextEncodingNone('some text here')

    """

    def __init__(self, size: Union[int, str]):
        pass

    def to_string(self) -> str:
        """
        Returns a Python string
        """
        pass


@extending.overload(operator.eq)
def text_encoding_none_eq(a, b):
    if isinstance(a, TextEncodingNonePointer) and isinstance(b, TextEncodingNonePointer):

        def impl(a, b):
            if len(a) != len(b):
                return False
            for i in range(0, len(a)):
                if a[i] != b[i]:
                    return False
            return True
        return impl
    elif isinstance(a, TextEncodingNonePointer) and isinstance(b, nb_types.StringLiteral):
        lv = b.literal_value
        sz = len(lv)

        def impl(a, b):
            if len(a) != sz:
                return False
            t = TextEncodingNone(lv)
            return a == t
        return impl


@extending.overload(operator.ne)
def text_encoding_none_ne(a, b):
    if isinstance(a, TextEncodingNonePointer):
        if isinstance(b, (TextEncodingNonePointer, nb_types.StringLiteral)):
            def impl(a, b):
                return not (a == b)
            return impl


@extending.lower_builtin(TextEncodingNone, nb_types.Integer)
def heavydb_text_encoding_none_constructor(context, builder, sig, args):
    return heavydb_buffer_constructor(context, builder, sig, args)


@extending.lower_builtin(TextEncodingNone, nb_types.CPointer, nb_types.Integer)
def heavydb_text_encoding_none_constructor_memcpy(context, builder, sig, args):
    [ptr, sz] = args
    fa = heavydb_buffer_constructor(context, builder, sig.return_type(nb_types.int64), [sz])
    fa_ptr = builder.load(builder.gep(fa, [i32(0), i32(0)]))
    cgutils.memcpy(builder, fa_ptr, ptr, sz)
    # string is null terminated
    builder.store(fa_ptr.type.pointee(0), builder.gep(fa_ptr, [sz]))
    return fa


@extending.lower_builtin(TextEncodingNone, nb_types.UnicodeType)
def text_encoding_none_unicode_ctor(context, builder, sig, args):
    # Note to the future:
    # You shall be tempted to use "context.make_helper",
    # "cgutils.make_struct_proxy" or "builder.extract_value", but avoid it at
    # all cost! LLVM -O2 runs SROA (Scalar Replacement Of Aggregates) and
    # suddenly some IR nodes becomes NULL values which are deleted by DCE.
    # Stick to plain allocas + load/store and you'll be fine.

    # It seems to be safe to use context.make_helper on a previous allocated
    # memory!
    uni_str = context.make_helper(builder, sig.args[0], value=args[0])

    llty = context.get_value_type(sig.return_type.dtype)
    st_size = context.get_abi_sizeof(llty)
    st_ptr = builder.bitcast(allocate_varlen_buffer(builder, int64_t(st_size),
                                                    int64_t(1)),
                             llty.as_pointer())
    zero, one, two = i32(0), i32(1), i32(2)

    eq = builder.icmp_signed('==', uni_str.length, uni_str.length.type(0))
    with builder.if_else(eq, likely=False) as (then, otherwise):
        with then:
            bb_then = builder.basic_block
            p1 = cgutils.get_null_value(i8p)
            n1 = i8(1)
        with otherwise:
            bb_else = builder.basic_block
            p2 = allocate_varlen_buffer(builder, uni_str.length, i64(1))
            cgutils.raw_memcpy(builder, dst=p2, src=uni_str.data,
                               count=uni_str.length, itemsize=i64(1))
            # string is null terminated
            builder.store(i8(0), builder.gep(p2, [uni_str.length]))
            n2 = i8(0)

    # {i8*, i64, i8}
    #  ^^^
    ptr = builder.phi(i8p)
    ptr.add_incoming(p1, bb_then)
    ptr.add_incoming(p2, bb_else)

    # {i8*, i64, i8}
    #            ^^
    is_null = builder.phi(i8)
    is_null.add_incoming(n1, bb_then)
    is_null.add_incoming(n2, bb_else)

    builder.store(ptr, builder.gep(st_ptr, [zero, zero]))
    builder.store(uni_str.length, builder.gep(st_ptr, [zero, one]))
    builder.store(is_null, builder.gep(st_ptr, [zero, two]))
    return st_ptr


@extending.type_callable(TextEncodingNone)
def type_heavydb_text_encoding_none(context):
    def typer(arg):
        if isinstance(arg, (nb_types.Integer, nb_types.UnicodeType)):
            return typesystem.Type.fromobject('TextEncodingNone').tonumba()
    return typer


@extending.type_callable(TextEncodingNone)
def type_heavydb_text_encoding_none_typer2(context):
    def typer(ptr, sz):
        if isinstance(ptr, nb_types.CPointer) and isinstance(sz, nb_types.Integer):
            return typesystem.Type.fromobject('TextEncodingNone').tonumba()
    return typer


@extending.intrinsic
def ol_attr_ptr_(typingctx, text):
    sig = nb_types.voidptr(text)

    def codegen(context, builder, sig, args):
        [data] = args
        return builder.extract_value(builder.load(data), 0)

    return sig, codegen


@extending.intrinsic
def ol_attr_sz_(typingctx, text):
    sig = nb_types.int64(text)

    def codegen(context, builder, sig, args):
        [data] = args
        return builder.extract_value(builder.load(data), 1)

    return sig, codegen


@extending.intrinsic
def ol_attr_is_null_(typingctx, text):
    sig = nb_types.int8(text)

    def codegen(context, builder, sig, args):
        [data] = args
        return builder.extract_value(builder.load(data), 2)

    return sig, codegen


@extending.overload_attribute(TextEncodingNonePointer, 'ptr')
def ol_attr_ptr(text):
    def impl(text):
        return ol_attr_ptr_(text)
    return impl


@extending.overload_attribute(TextEncodingNonePointer, 'sz')
def ol_attr_sz(text):
    def impl(text):
        return ol_attr_sz_(text)
    return impl


@extending.overload_attribute(TextEncodingNonePointer, 'is_null')
def ol_attr_is_null(text):
    def impl(text):
        return ol_attr_is_null_(text)
    return impl


@extending.overload_method(TextEncodingNonePointer, 'to_string')
def ol_to_string(text):
    def impl(text):
        kind = PY_UNICODE_1BYTE_KIND  # ASCII characters only
        length = len(text)
        is_ascii = True
        s = _empty_string(kind, length, is_ascii)
        for i in range(length):
            ch = text[i]
            _set_code_point(s, i, ch)
        return s
    return impl



import pytest
from rbc.irtools import read_unicodetype_db


@pytest.fixture(scope='module')
def db():
    return read_unicodetype_db()


@pytest.mark.usefixtures('db')
def test_unicodedbtype_functions(db):
    # ensure there's a unicodetype_db.ll file and it contains the
    # required symbols
    required_fns = ('numba_PyUnicode_ToNumeric', 'numba_PyUnicode_IsWhitespace',
                    'numba_PyUnicode_IsLinebreak', 'numba_gettyperecord',
                    'numba_get_PyUnicode_ExtendedCase')
    fns = list(db.functions)
    assert len(fns) == len(required_fns)
    for fn in db.functions:
        assert fn.name in required_fns


@pytest.mark.usefixtures('db')
def test_unicodedbtype_globals(db):

    required_globals = ('numba_PyUnicode_TypeRecords', 'index1', 'index2',
                        'numba_PyUnicode_ExtendedCase')
    for glob in db.global_variables:
        assert glob.name in required_globals


@pytest.mark.usefixtures('db')
def test_numba_PyUnicode_TypeRecord_struct(db):

    try:
        db.get_struct_type('struct.numba_PyUnicode_TypeRecord')
    except NameError:
        pytest.fail('Raise exception')



import atexit
import pytest
from rbc.remotejit import RemoteJIT
import numpy as np


@pytest.fixture(scope="module")
def rjit(request):
    rjit = RemoteJIT(port=11541)
    rjit.start_server(background=True)
    request.addfinalizer(rjit.stop_server)
    atexit.register(rjit.stop_server)
    return rjit


def test_trunc(rjit):
    @rjit('double(double)')
    def trunc(x):
        return np.trunc(x)

    assert (np.allclose(trunc(0.3), np.trunc(0.3)))
    assert (np.allclose(trunc(2.9), np.trunc(2.0)))


def test_exp2(rjit):
    @rjit('double(double)')
    def exp2(x):
        return np.exp2(x)

    assert (np.isclose(exp2(2), np.exp2(2)))
    assert (np.isclose(exp2(3), np.exp2(3)))


def test_exp2f(rjit):
    @rjit('float(float)')
    def exp2(x):
        return np.exp2(x)

    assert (np.isclose(exp2(2.0), np.exp2(2.0)))


def test_logaddexp(rjit):
    # ref:
    # https://github.com/numpy/numpy/blob/2fd9ff8277ad25aa386c3432b6ebc35322879d91/numpy/core/tests/test_umath.py#L818-L860

    @rjit('double(double, double)')
    def logaddexp(x, y):
        return np.logaddexp(x, y)

    def test_values():
        x = [1, 2, 3, 4, 5]
        y = [5, 4, 3, 2, 1]
        z = [6, 6, 6, 6, 6]
        for _x, _y, _z in zip(x, y, z):
            for dt in ['f', 'd', 'g']:
                logxf = np.log(np.array(_x, dtype=dt))
                logyf = np.log(np.array(_y, dtype=dt))
                logzf = np.log(np.array(_z, dtype=dt))
                assert np.allclose(logaddexp(logxf, logyf), logzf)

    def test_range():
        x = [1000000, -1000000, 1000200, -1000200]
        y = [1000200, -1000200, 1000000, -1000000]
        z = [1000200, -1000000, 1000200, -1000000]
        for _x, _y, _z in zip(x, y, z):
            for dt in ['f', 'd', 'g']:
                # [()] ~> quick hack! rbc doesn't support passing arrays
                xf = np.array(_x, dtype=dt)[()]
                yf = np.array(_y, dtype=dt)[()]
                zf = np.array(_z, dtype=dt)[()]
                assert np.allclose(logaddexp(xf, yf), zf)

    def test_inf():
        # logaddexp inf
        inf = np.inf
        x = [inf, -inf,  inf, -inf, inf, 1,  -inf,  1]
        y = [inf,  inf, -inf, -inf, 1,   inf, 1,   -inf]
        z = [inf,  inf,  inf, -inf, inf, inf, 1,    1]
        with np.errstate(invalid='raise'):
            for _x, _y, _z in zip(x, y, z):
                for dt in ['f', 'd', 'g']:
                    # [()] ~> quick hack! rbc doesn't support passing arrays
                    xf = np.array(_x, dtype=dt)[()]
                    yf = np.array(_y, dtype=dt)[()]
                    zf = np.array(_z, dtype=dt)[()]
                    assert np.allclose(logaddexp(xf, yf), zf)

    def test_nan():
        assert np.isnan(logaddexp(np.nan, np.inf))
        assert np.isnan(logaddexp(np.inf, np.nan))
        assert np.isnan(logaddexp(np.nan, 0))
        assert np.isnan(logaddexp(0, np.nan))
        assert np.isnan(logaddexp(np.nan, np.nan))

    test_values()
    test_range()
    test_inf()
    test_nan()


def test_logaddexp2(rjit):
    # ref:
    # https://github.com/numpy/numpy/blob/2fd9ff8277ad25aa386c3432b6ebc35322879d91/numpy/core/tests/test_umath.py#L580-L619
    @rjit('double(double, double)')
    def logaddexp2(x, y):
        return np.logaddexp2(x, y)

    def test_values():
        x = [1, 2, 3, 4, 5]
        y = [5, 4, 3, 2, 1]
        z = [6, 6, 6, 6, 6]
        for _x, _y, _z in zip(x, y, z):
            for dt in ['f', 'd', 'g']:
                logxf = np.log2(np.array(_x, dtype=dt))
                logyf = np.log2(np.array(_y, dtype=dt))
                logzf = np.log2(np.array(_z, dtype=dt))
                assert np.allclose(logaddexp2(logxf, logyf), logzf)

    def test_range():
        x = [1000000, -1000000, 1000200, -1000200]
        y = [1000200, -1000200, 1000000, -1000000]
        z = [1000200, -1000000, 1000200, -1000000]
        for _x, _y, _z in zip(x, y, z):
            for dt in ['f', 'd', 'g']:
                # [()] ~> quick hack! rbc doesn't support passing arrays
                xf = np.array(_x, dtype=dt)[()]
                yf = np.array(_y, dtype=dt)[()]
                zf = np.array(_z, dtype=dt)[()]
                assert np.allclose(logaddexp2(xf, yf), zf)

    def test_inf():
        # logaddexp2 inf
        inf = np.inf
        x = [inf, -inf,  inf, -inf, inf, 1,  -inf,  1]
        y = [inf,  inf, -inf, -inf, 1,   inf, 1,   -inf]
        z = [inf,  inf,  inf, -inf, inf, inf, 1,    1]
        with np.errstate(invalid='raise'):
            for _x, _y, _z in zip(x, y, z):
                for dt in ['f', 'd', 'g']:
                    # [()] ~> quick hack! rbc doesn't support passing arrays
                    xf = np.array(_x, dtype=dt)[()]
                    yf = np.array(_y, dtype=dt)[()]
                    zf = np.array(_z, dtype=dt)[()]
                    assert np.allclose(logaddexp2(xf, yf), zf)

    def test_nan():
        assert np.isnan(logaddexp2(np.nan, np.inf))
        assert np.isnan(logaddexp2(np.inf, np.nan))
        assert np.isnan(logaddexp2(np.nan, 0))
        assert np.isnan(logaddexp2(0, np.nan))
        assert np.isnan(logaddexp2(np.nan, np.nan))

    test_values()
    test_range()
    test_inf()
    test_nan()


def test_signbit(rjit):

    @rjit('bool(double)')
    def signbit(x):
        return np.signbit(x)

    def check(array):
        for val in array:
            out = signbit(val)
            expected = np.signbit(val)
            assert out == expected, val

    a = np.arange(-10, 10, dtype=np.float32)
    check(a)

    b = np.array([-2, 5, 1, 4, 3], dtype=np.float16)
    check(b)

    c = np.array([-0.0, 0.0, np.inf, -np.inf], dtype=np.float32)
    check(c)



"""
Array API specification for linear algebra functions.

https://data-apis.org/array-api/latest/API_specification/linear_algebra_functions.html
"""
from rbc.stdlib import Expose

__all__ = ["matmul", "matrix_transpose", "tensordot", "vecdot"]

expose = Expose(globals(), "linear_algebra_functions")


@expose.not_implemented("matmul")
def _array_api_matmul(x1, x2):
    """
    Computes the matrix product.
    """
    pass


@expose.not_implemented("matrix_transpose")
def _array_api_matrix_transpose(x):
    """
    Transposes a matrix (or a stack of matrices) x.
    """
    pass


@expose.not_implemented("tensordot")
def _array_api_tensordot(x1, x2, *, axes=2):
    """
    Returns a tensor contraction of x1 and x2 over specific axis.
    """
    pass


@expose.not_implemented("vecdot")
def _array_api_vecdot(x1, x2, *, axis=-1):
    """
    Computes the (vector) dot product of two arrays.
    """
    pass



import pytest
from rbc.tests import heavydb_fixture

ARRAY_NULL = 0
ARRAY_IDX_IS_NULL = 1
ARRAY_NOT_NULL = 2


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(5, 6), suffices=['arraynull']):
        define(o)
        yield o


def define(heavydb):

    @heavydb('int64(T[], int64)', T=['bool', 'int8', 'int16', 'int32', 'int64', 'float', 'double'])
    def array_null_check(x, index):
        if x.is_null():  # array row is null
            return ARRAY_NULL
        if x.is_null(index):  # array index is null
            return ARRAY_IDX_IS_NULL
        return ARRAY_NOT_NULL


colnames = ['b', 'i1', 'i2', 'i4', 'i8', 'f4', 'f8']


@pytest.mark.parametrize('col', colnames)
def test_array_null(heavydb, col):
    if col in ['i2', 'i8', 'f8']:
        heavydb.require_version((5, 7, 0),
                                'Requires heavydb-internal PR 5465 [rbc PR 330]')
    if col == 'b':
        heavydb.require_version((5, 7, 0),
                                'Requires heavydb-internal PR 5492 [rbc issue 245]')

    # Query null value
    _, result = heavydb.sql_execute(f'''
        SELECT
            array_null_check({col}, 0)
        FROM
            {heavydb.table_name}arraynull
    ''')
    result = list(result)

    expected = {
        'b':  [(ARRAY_NULL,), (ARRAY_NOT_NULL,), (ARRAY_NULL,),
               (ARRAY_IDX_IS_NULL,), (ARRAY_NULL,)],
        'i1': [(ARRAY_NULL,), (ARRAY_IDX_IS_NULL,), (ARRAY_NULL,),
               (ARRAY_NOT_NULL,), (ARRAY_NULL,)],
        'i2': [(ARRAY_NOT_NULL,), (ARRAY_NULL,), (ARRAY_NOT_NULL,),
               (ARRAY_NULL,), (ARRAY_NOT_NULL,)],
        'i4': [(ARRAY_NULL,), (ARRAY_NOT_NULL,), (ARRAY_NULL,),
               (ARRAY_NOT_NULL,), (ARRAY_NULL,)],
        'i8': [(ARRAY_NOT_NULL,), (ARRAY_NULL,), (ARRAY_NOT_NULL,),
               (ARRAY_NULL,), (ARRAY_IDX_IS_NULL,)],
        'f4': [(ARRAY_NULL,), (ARRAY_NOT_NULL,), (ARRAY_NULL,),
               (ARRAY_IDX_IS_NULL,), (ARRAY_NULL,)],
        'f8': [(ARRAY_IDX_IS_NULL,), (ARRAY_NULL,), (ARRAY_IDX_IS_NULL,),
               (ARRAY_NULL,), (ARRAY_NOT_NULL,)],
    }

    assert result == expected[col]




__all__ = ['StructureNumbaType', 'StructureNumbaPointerType', 'make_numba_struct']

import operator
from llvmlite import ir
from numba.core import datamodel, extending, types, imputils, typing, cgutils, typeconv


typing_registry = typing.templates.builtin_registry
lowering_registry = imputils.builtin_registry

int8_t = ir.IntType(8)
int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
void_t = ir.VoidType()
fp32 = ir.FloatType()
fp64 = ir.DoubleType()


class StructureTypeAttributeTemplate(typing.templates.AttributeTemplate):
    key = NotImplemented

    def generic_resolve(self, typ, attr):
        model = datamodel.default_manager.lookup(typ)
        try:
            return model.get_member_fe_type(attr)
        except KeyError:
            pass


class StructurePointerTypeAttributeTemplate(typing.templates.AttributeTemplate):
    key = NotImplemented

    def generic_resolve(self, typ, attr):
        model = datamodel.default_manager.lookup(typ.dtype)
        return model.get_member_fe_type(attr)


class StructureNumbaType(types.Type):
    """Represents a struct numba type.
    """


class StructureNumbaPointerType(types.Type):
    """Pointer type for StructureNumbaType values.

    We are not deriving from CPointer because we may want to use
    getitem for custom item access.
    """

    @property
    def __typesystem_type__(self):
        return Type.fromnumba(self.dtype).pointer()

    def __init__(self, dtype):
        self.dtype = dtype    # struct dtype
        name = "%s[%s]*" % (type(self).__name__, dtype)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype

    # Subclasses may redefine the following methods

    def get_add_impl(self):
        def impl(x, i):
            return StructureNumbaPointerType_add_impl(x, i)
        return impl

    def get_getitem_impl(self):
        return self.get_add_impl()


def make_numba_struct(name, members, origin=None, _cache={}):
    """Create numba struct type instance.
    """
    t = _cache.get(name)
    if t is None:
        base = None
        if origin is not None:
            assert origin.is_struct
            base = origin._params.get('NumbaType')
            numba_ptr_type = origin._params.get('NumbaPointerType', StructureNumbaPointerType)
        else:
            base = None
            numba_ptr_type = StructureNumbaPointerType
        if base is None:
            base = Type.aliases.get('StructureNumbaType', StructureNumbaType)
        assert issubclass(base, types.Type)  # base must be numba.types.Type
        assert issubclass(numba_ptr_type, types.Type)  # numba_ptr_type must be numba.types.Type

        struct_type = type(name+'Type', (base,),
                           dict(members=[t for n, t in members],
                                __typesystem_type__=origin))
        _cache[name] = t = struct_type(name)

        def model__init__(self, dmm, fe_type):
            datamodel.StructModel.__init__(self, dmm, fe_type, members)
        struct_model = type(name+'Model',
                            (datamodel.StructModel,),
                            dict(__init__=model__init__))
        struct_pointer_model = type(name+'PointerModel', (datamodel.models.PointerModel,), dict())

        datamodel.registry.register_default(struct_type)(struct_model)
        datamodel.registry.register_default(numba_ptr_type)(struct_pointer_model)

        if not issubclass(numba_ptr_type, StructureNumbaPointerType):
            return t

        # member access via attribute getters/setters
        struct_attr_template = type(name+'AttributeTemplate',
                                    (StructureTypeAttributeTemplate,),
                                    dict(key=struct_type))
        struct_pointer_attr_template = type(name+'PointerAttributeTemplate',
                                            (StructurePointerTypeAttributeTemplate,),
                                            dict(key=numba_ptr_type))
        typing_registry.register_attr(struct_pointer_attr_template)
        typing_registry.register_attr(struct_attr_template)

        # allow casting to/from voidptr
        tptr = numba_ptr_type(t)
        typeconv.rules.default_type_manager.set_compatible(
            tptr, types.voidptr, typeconv.Conversion.safe)
        typeconv.rules.default_type_manager.set_compatible(
            types.voidptr, tptr, typeconv.Conversion.safe)

    return t


@lowering_registry.lower_getattr_generic(StructureNumbaType)
def StructureNumbaType_getattr_impl(context, builder, sig, struct, attr):
    model = datamodel.default_manager.lookup(sig)
    index = model.get_field_position(attr)
    return builder.extract_value(struct, [index])


@lowering_registry.lower_setattr_generic(StructureNumbaType)
def StructureNumbaType_setattr_impl(context, builder, sig, args, attr):
    typ = sig.args[0]
    struct, value = args
    model = datamodel.default_manager.lookup(typ)
    index = model.get_field_position(attr)
    ptr = builder.gep(struct.operands[0], [int32_t(0), int32_t(index)])
    return builder.store(value, ptr)


@lowering_registry.lower_getattr_generic(StructureNumbaPointerType)
def StructureNumbaPointerType_getattr_impl(context, builder, sig, struct, attr):
    typ = sig.dtype
    model = datamodel.default_manager.lookup(typ)
    index = model.get_field_position(attr)
    rawptr = cgutils.alloca_once_value(builder, value=struct)
    struct = builder.load(builder.gep(rawptr, [int32_t(0)]))
    return builder.load(builder.gep(
        struct, [int32_t(0), int32_t(index)]))


@lowering_registry.lower_setattr_generic(StructureNumbaPointerType)
def StructureNumbaPointerType_setattr_impl(context, builder, sig, args, attr):
    typ = sig.args[0].dtype
    struct, value = args
    model = datamodel.default_manager.lookup(typ)
    index = model.get_field_position(attr)
    rawptr = cgutils.alloca_once_value(builder, value=struct)
    ptr = builder.load(rawptr)
    buf = builder.load(builder.gep(ptr, [int32_t(0), int32_t(index)]))
    builder.store(value, buf.operands[0])


@extending.intrinsic
def StructureNumbaPointerType_add_impl(typingctx, data, index):
    sig = data(data, index)
    ptr_type = data

    def codegen(context, builder, signature, args):
        ll_ptr_type = context.get_value_type(ptr_type)
        ll_value_type = context.get_value_type(ptr_type.dtype)
        size = context.get_abi_sizeof(ll_value_type)
        ptr, index = args

        i = builder.ptrtoint(ptr, int64_t)
        offset = builder.mul(index, int64_t(size))
        new_i = builder.add(i, offset)

        return builder.inttoptr(new_i, ll_ptr_type)
    return sig, codegen


@extending.overload(operator.add)
def StructureNumbaPointerType_add(x, i):
    if isinstance(x, StructureNumbaPointerType):
        return x.get_add_impl()


@extending.overload(operator.getitem)
def StructureNumbaPointerType_getitem(x, i):
    if isinstance(x, StructureNumbaPointerType):
        return x.get_getitem_impl()


@lowering_registry.lower_cast(types.RawPointer, StructureNumbaPointerType)
@lowering_registry.lower_cast(StructureNumbaPointerType, types.RawPointer)
def impl_T_star_to_T_star(context, builder, fromty, toty, value):
    return builder.bitcast(value, Type.fromnumba(toty).tollvmir())


if 1:
    from rbc.typesystem import Type



import os
import itertools
import pytest
import numpy as np

from rbc.errors import UnsupportedError, HeavyDBServerError
from rbc.tests import heavydb_fixture, assert_equal
from rbc.typesystem import Type

rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.mark.parametrize("heavydb_brand", ['heavyai'])
def test_get_client_config(tmpdir, heavydb_brand):
    d = tmpdir.mkdir(heavydb_brand)
    fh = d.join("client.conf")
    fh.write(f"""
    [user]
name  =  foo
password = secret

[server]
port: 1234
host: example.com
dbname: {heavydb_brand}

[rbc]
debug: False
use_host_target: False
# server: Server [NOT IMPL]
# target_info: TargetInfo
""")
    conf_file = os.path.join(fh.dirname, fh.basename)

    client_conf_env = dict(heavyai='HEAVYDB_CLIENT_CONF')[heavydb_brand]
    old_conf = os.environ.get(client_conf_env)
    os.environ[client_conf_env] = conf_file

    try:
        conf = rbc_heavydb.get_client_config()
        assert conf['user'] == 'foo'
        assert conf['password'] == 'secret'
        assert conf['port'] == 1234
        assert conf['host'] == 'example.com'
        assert conf['dbname'] == heavydb_brand
        assert conf['debug'] == bool(0)
        conf = rbc_heavydb.get_client_config(dbname='test')
        assert conf['dbname'] == 'test'
    finally:
        if old_conf is None:
            del os.environ[client_conf_env]
        else:
            os.environ[client_conf_env] = old_conf


@pytest.fixture(scope='module')
def nb_version():
    from rbc.utils import get_version
    return get_version('numba')


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), suffices=['']):
        yield o


def test_direct_call(heavydb):
    heavydb.reset()

    @heavydb('double(double)')
    def farhenheit2celcius(f):
        return (f - 32) * 5 / 9

    assert_equal(farhenheit2celcius(40).execute(), np.float32(40 / 9))


def test_local_caller(heavydb):
    heavydb.reset()

    def func(f):
        return f

    caller = heavydb('double(double)')(func)

    msg = "Cannot create a local `Caller`"
    with pytest.raises(UnsupportedError, match=msg):
        _ = caller.local


def test_redefine(heavydb):
    heavydb.reset()

    @heavydb('i32(i32)')
    def incr(x):
        return x + 1

    desrc, result = heavydb.sql_execute(
        f'select i4, incr(i4) from {heavydb.table_name}')
    for x, x1 in result:
        assert x1 == x + 1

    # Re-defining triggers a warning message when in debug mode
    @heavydb('i32(i32)')  # noqa: F811
    def incr(x):  # noqa: F811
        return x + 2

    desrc, result = heavydb.sql_execute(
        f'select i4, incr(i4) from {heavydb.table_name}')
    for x, x1 in result:
        assert x1 == x + 2


def test_single_argument_overloading(heavydb):
    heavydb.reset()

    @heavydb(
        'f64(f64)',
        'i64(i64)',
        'i32(i32)',
        'f32(f32)',
        # 'i32(f32)',
    )
    def mydecr(x):
        return x - 1
    desrc, result = heavydb.sql_execute(
        'select f4, mydecr(f4) from {heavydb.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for x, x1 in result:
        assert x1 == x - 1
        assert isinstance(x1, type(x))
    desrc, result = heavydb.sql_execute(
        'select f8, mydecr(f8) from {heavydb.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for x, x1 in result:
        assert x1 == x - 1
        assert isinstance(x1, type(x))
    desrc, result = heavydb.sql_execute(
        'select i4, mydecr(i4) from {heavydb.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for x, x1 in result:
        assert x1 == x - 1
        assert isinstance(x1, type(x))


def test_thrift_api_doc(heavydb):
    heavydb.reset()

    @heavydb('double(int, double)',
             'float(int, float)',
             'int(int, int)')
    def foo(i, v):
        return v * i + 55

    descr, result = heavydb.sql_execute(
        'select f8, foo(i4, f8) from {heavydb.table_name}'.format(**locals()))
    result = list(result)
    assert len(result) > 0
    for i, (x, x1) in enumerate(result):
        assert x1 == x * i + 55
        assert isinstance(x1, type(x))


def test_multiple_implementation(heavydb):
    heavydb.reset()

    @heavydb('int(f64)', 'int(i64)')  # noqa: F811
    def bits(x):
        return 64

    @heavydb('int(f32)', 'int(i32)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 32

    @heavydb('int(i16)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 16

    @heavydb('int(i8)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 8

    descr, result = heavydb.sql_execute(
        'select bits(i1), bits(i2), bits(i4), bits(f4), bits(i8), bits(f8)'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    result = list(result)
    assert len(result) == 1
    assert result[0] == (8, 16, 32, 32, 64, 64)


def test_loadtime_udf(heavydb):
    # This test requires that heavydb server is started with `--udf
    # sample_udf.cpp` option where the .cpp file defines `udf_diff`
    try:
        desrc, result = heavydb.sql_execute(
            'select i4, udf_diff2(i4, i4) from {heavydb.table_name}'
            .format(**locals()))
    except Exception as msg:
        assert "Undefined function call 'udf_diff2'" in str(msg)
        return
    result = list(result)
    for i_, (i, d) in enumerate(result):
        assert i_ == i
        assert d == 0


def test_f32(heavydb):
    """If UDF name ends with an underscore, expect strange behaviour. For
    instance, defining

      @heavydb('f32(f32)', 'f32(f64)')
      def f32_(x): return x+4.5

    the query `select f32_(0.0E0))` fails but not when defining

      @heavydb('f32(f64)', 'f32(f32)')
      def f32_(x): return x+4.5

    (notice the order of signatures in heavydb decorator argument).
    """

    @heavydb('f32(f32)', 'f32(f64)')  # noqa: F811
    def f_32(x): return x+4.5
    descr, result = heavydb.sql_execute(
        'select f_32(0.0E0) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5,)


def test_castop(heavydb):
    @heavydb('i16(i16)')  # noqa: F811
    def i32(x): return x+2  # noqa: F811

    @heavydb('i32(i32)')  # noqa: F811
    def i32(x): return x+4  # noqa: F811

    descr, result = heavydb.sql_execute(
        'select i32(cast(1 as int))'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (5,)


def test_binding(heavydb):
    heavydb.reset()
    full = os.environ.get('RBC_TESTS_FULL', 'FALSE').lower() in ['1', 'true', 'on']

    if full:
        argument_types = ['i8', 'i16', 'i32', 'i64', 'f32', 'f64']
        literals = ['0', '0.0', 'cast(0 as tinyint)', 'cast(0 as smallint)', 'cast(0 as int)',
                    'cast(0 as bigint)', 'cast(0 as float)', 'cast(0 as double)']
        literals_types = ['i32', 'f32'] + argument_types
        column_vars = ['cast(i8 as tinyint)', 'cast(i8 as smallint)', 'cast(i8 as int)',
                       'cast(i8 as bigint)', 'cast(i8 as float)', 'cast(i8 as double)']
        column_vars_types = argument_types
    else:
        # skip smallint and int
        argument_types = ['i8', 'i64', 'f32', 'f64']
        literals = ['0', '0.0', 'cast(0 as tinyint)', 'cast(0 as bigint)',
                    'cast(0 as float)', 'cast(0 as double)']
        literals_types = ['i32', 'f32'] + argument_types
        column_vars = ['cast(i8 as tinyint)', 'cast(i8 as bigint)',
                       'cast(i8 as float)', 'cast(i8 as double)']
        column_vars_types = argument_types

    if available_version[:2] >= (5, 9):
        heavydb.require_version((5, 9), 'Requires heavydb-internal PR 6003')

        def get_result(overload_types, input_type, is_literal):
            overload_types_ = overload_types[::-1 if is_literal else 1]
            for overload_type in overload_types_:
                if input_type[0] == overload_type[0]:
                    if int(input_type[1:]) <= int(overload_type[1:]):
                        return overload_type
            if input_type[0] == 'i':
                for overload_type in reversed(overload_types):
                    if overload_type[0] == 'f':
                        return overload_type
            return 'NO BINDING FOUND'

    elif available_version[:2] == (5, 8):

        def get_result(overload_types, input_type, is_literal):
            overload_types_ = overload_types[::-1 if is_literal else 1]
            for overload_type in overload_types_:
                if input_type[0] == overload_type[0]:
                    if int(input_type[1:]) <= int(overload_type[1:]):
                        return overload_type
            if input_type[0] == 'i':
                for overload_type in overload_types_:
                    if overload_type[0] == 'f' and int(input_type[1:]) <= int(overload_type[1:]):
                        return overload_type
            return 'NO BINDING FOUND'

    else:
        # typeless literal expressions not support as arguments
        literals = literals[2:]
        literals_types = literals_types[2:]

        def get_result(overload_types, input_type, is_literal):
            for overload_type in overload_types:
                if input_type[0] == overload_type[0]:
                    if int(input_type[1:]) <= int(overload_type[1:]):
                        return overload_type
                elif overload_type == 'f64':
                    return overload_type
            return 'NO BINDING FOUND'

    assert len(literals) == len(literals_types)
    assert len(column_vars) == len(column_vars_types)

    heavydb.reset()
    functions = []
    for r in range(1, len(argument_types) + 1):
        if not full and r not in [1, 2, len(argument_types)]:
            continue
        for p in itertools.combinations(range(len(argument_types)), r=r):
            fname = 'rt_binding_' + '_'.join(map(argument_types.__getitem__, p))
            for i in p:
                foo = eval(f'lambda x: {i}')
                foo.__name__ = fname
                heavydb(f'i8({argument_types[i]})')(foo)
            functions.append(fname)
    heavydb.register()

    for fname in functions:
        for input, input_type in zip(literals, literals_types):
            try:
                descr, result = heavydb.sql_execute(f'select {fname}({input})')
                result = list(result)
                result = argument_types[result[0][0]]
            except Exception:
                result = 'NO BINDING FOUND'
            expected = get_result(fname[len('rt_binding_'):].split('_'), input_type, True)
            assert result == expected, (fname, input, input_type, result, expected)

        for input, input_type in zip(column_vars, column_vars_types):
            try:
                descr, result = heavydb.sql_execute(
                    f'select {fname}({input}) from {heavydb.table_name} limit 1')
                result = list(result)
                result = argument_types[result[0][0]]
            except Exception:
                result = 'NO BINDING FOUND'
            expected = get_result(fname[len('rt_binding_'):].split('_'), input_type, False)
            assert result == expected, (fname, input, input_type, result, expected)


def test_casting(heavydb):
    """Define UDFs:
      i8(<tinyint>), i16(<smallint>), i32(<int>), i64(<bigint>)
      f32(<float>), f64(<double>)

    The following table defines the behavior of applying these UDFs to
    values with different types:

    OmnisciDB version 5.9+
    ----------------------
             | Functions applied to <itype value>
    itype    | i8   | i16  | i32  | i64  | f32  | f64  |
    ---------+------+------+------+------+------+------+
    tinyint  | OK   | OK   | OK   | OK   | OK   | OK   |
    smallint | FAIL | OK   | OK   | OK   | OK   | OK   |
    int      | FAIL | FAIL | OK   | OK   | OK   | OK   |
    bigint   | FAIL | FAIL | FAIL | OK   | OK   | OK   |
    float    | FAIL | FAIL | FAIL | FAIL | OK   | OK   |
    double   | FAIL | FAIL | FAIL | FAIL | FAIL | OK   |

    OmnisciDB version 5.8
    ----------------------
             | Functions applied to <itype value>
    itype    | i8   | i16  | i32  | i64  | f32  | f64  |
    ---------+------+------+------+------+------+------+
    tinyint  | OK   | OK   | OK   | OK   | OK   | OK   |
    smallint | FAIL | OK   | OK   | OK   | OK   | OK   |
    int      | FAIL | FAIL | OK   | OK   | OK   | OK   |
    bigint   | FAIL | FAIL | FAIL | OK   | FAIL | OK   |
    float    | FAIL | FAIL | FAIL | FAIL | OK   | OK   |
    double   | FAIL | FAIL | FAIL | FAIL | FAIL | OK   |

    OmnisciDB version 5.7 and older
    -------------------------------
             | Functions applied to <itype value>
    itype    | i8   | i16  | i32  | i64  | f32  | f64  |
    ---------+------+------+------+------+------+------+
    tinyint  | OK   | OK   | OK   | OK   | FAIL | FAIL |
    smallint | FAIL | OK   | OK   | OK   | FAIL | FAIL |
    int      | FAIL | FAIL | OK   | OK   | FAIL | FAIL |
    bigint   | FAIL | FAIL | FAIL | OK   | FAIL | FAIL |
    float    | FAIL | FAIL | FAIL | FAIL | OK   | OK   |
    double   | FAIL | FAIL | FAIL | FAIL | FAIL | OK   |

    test_binding is superior test with respect to successful UDF
    executions but it does not check exception messages.
    """
    heavydb.reset()

    @heavydb('i8(i8)')    # noqa: F811
    def i8(x): return x+1

    @heavydb('i16(i16)')  # noqa: F811
    def i16(x): return x+2

    @heavydb('i32(i32)')  # noqa: F811
    def i32(x): return x+4

    @heavydb('i64(i64)')  # noqa: F811
    def i64(x): return x+8

    @heavydb('f32(f32)')  # noqa: F811
    def f32(x): return x+4.5

    # cannot create a 8-bit and 16-bit int literals in sql, so, using
    # the following helper functions:
    @heavydb('i8(i8)', 'i8(i16)', 'i8(i32)')    # noqa: F811
    def i_8(x): return x+1

    @heavydb('i16(i16)', 'i16(i32)')  # noqa: F811
    def i_16(x): return x+2

    # cannot create a 32-bit float literal in sql, so, using a helper
    # function for that:
    @heavydb('f32(f32)', 'f32(f64)')  # noqa: F811
    def f_32(x): return x+4.5

    @heavydb('f64(f64)')  # noqa: F811
    def f64(x): return x+8.5

    @heavydb('i8(i8)')
    def ifoo(x): return x + 1

    @heavydb('i16(i16)')  # noqa: F811
    def ifoo(x): return x + 2  # noqa: F811

    @heavydb('i32(i32)')  # noqa: F811
    def ifoo(x): return x + 4  # noqa: F811

    @heavydb('i64(i64)')  # noqa: F811
    def ifoo(x): return x + 8  # noqa: F811

    @heavydb('f32(f32)')
    def ffoo(x): return x + 4.5  # noqa: F811

    @heavydb('f64(f64)')  # noqa: F811
    def ffoo(x): return x + 8.5  # noqa: F811

    rows = []
    for itype, ivalue in [('tinyint', 'i_8(0)'),
                          ('smallint', 'i_16(0)'),
                          ('int', 'i32(0)'),
                          ('bigint', 'i64(0)'),
                          ('float', 'f_32(0.0)'),
                          ('double', 'f64(0.0)')]:
        row = [f'{itype:8}']
        cols = [f'{"itype":8}']
        for atype, func in [('tinyint', 'i8'),
                            ('smallint', 'i16'),
                            ('int', 'i32'),
                            ('bigint', 'i64'),
                            ('float', 'f32'),
                            ('double', 'f64')]:
            cols.append(f'{func:4}')
            try:
                descr, result = heavydb.sql_execute(
                    f'select {func}({ivalue}) from {heavydb.table_name} limit 1')
                status = 'OK'
            except Exception:
                status = 'FAIL'
            row.append(f'{status:4}')
        if not rows:
            rows.append(' | '.join(cols))
            rows.append('+'.join([f'{"":-^9}'] + [f'{"":-^6}'] * (len(cols)-1)))
        rows.append(' | '.join(row))

    print('\n\nSUPPORTED CASTING RULES FOR SCALAR ARGUMENTS:\n')
    print('\n'.join(rows))

    descr, result = heavydb.sql_execute(
        'select i_8(0),i_16(0),i32(0),i64(0) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1, 2, 4, 8)

    descr, result = heavydb.sql_execute(
        'select i_8(i1),i16(i2),i32(i4),i64(i8)'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1, 2, 4, 8)

    descr, result = heavydb.sql_execute(
        'select ifoo(i_8(0)),ifoo(i_16(0)),ifoo(i32(0)),ifoo(i64(0))'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1+1, 2+2, 4+4, 8+8)

    descr, result = heavydb.sql_execute(
        'select ifoo(i1),ifoo(i2),ifoo(i4),ifoo(i8)'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (1, 2, 4, 8)

    descr, result = heavydb.sql_execute(
        'select i64(i_8(0)), i64(i_16(0)),i64(i32(0)),i64(i64(0))'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (9, 10, 12, 16)

    descr, result = heavydb.sql_execute(
        'select i64(i1), i64(i2),i64(i4),i64(i8)'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (8, 8, 8, 8)

    descr, result = heavydb.sql_execute(
        'select i32(i_8(0)), i32(i_16(0)),i32(i32(0))'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (5, 6, 8)

    descr, result = heavydb.sql_execute(
        'select i32(i1), i32(i2),i32(i4)'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (4, 4, 4)

    descr, result = heavydb.sql_execute(
        'select i16(i_8(0)), i16(i_16(0)) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (3, 4)

    descr, result = heavydb.sql_execute(
        'select i16(i1), i16(i2) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (2, 2)

    descr, result = heavydb.sql_execute(
        'select i8(i_8(0)) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (2,)

    descr, result = heavydb.sql_execute(
        'select i8(i1) from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (1,)

    descr, result = heavydb.sql_execute(
        'select f_32(0.0E0), f64(0.0E0) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5, 8.5)

    descr, result = heavydb.sql_execute(
        'select f_32(f4), f64(f8) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5, 8.5)

    descr, result = heavydb.sql_execute(
        'select f64(f_32(0.0E0)), f64(f64(0.0E0))'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (13.0, 17.0)

    descr, result = heavydb.sql_execute(
        'select f64(f4), f64(f8)'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (8.5, 8.5)

    descr, result = heavydb.sql_execute(
        'select f32(f_32(0.0E0)) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (9.0,)

    descr, result = heavydb.sql_execute(
        'select f32(f4) from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (4.5,)

    descr, result = heavydb.sql_execute(
        'select ffoo(f_32(0.0E0)), ffoo(f64(0.0E0))'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (9.0, 17.0)

    descr, result = heavydb.sql_execute(
        'select ffoo(f4), ffoo(f8)'
        ' from {heavydb.table_name} limit 1'.format(**locals()))
    assert list(result)[0] == (4.5, 8.5)

    for v, f, t in [
            ('i64(0)', r'i32', r'BIGINT'),
            ('i64(0)', r'i16', r'BIGINT'),
            ('i64(0)', r'i8', r'BIGINT'),
            ('i8', r'i32', r'BIGINT'),
            ('i8', r'i16', r'BIGINT'),
            ('i8', r'i8', r'BIGINT'),
            ('i32(0)', r'i16', r'INTEGER'),
            ('i32(0)', r'i8', r'INTEGER'),
            ('i4', r'i16', r'INTEGER'),
            ('i4', r'i8', r'INTEGER'),
            ('i_16(0)', r'i8', r'SMALLINT'),
            ('i2', r'i8', r'SMALLINT'),
            ('f64(0)', r'f32', r'DOUBLE'),
            ('f8', r'f32', r'DOUBLE'),
    ]:
        q = f'select {f}({v}) from {heavydb.table_name} limit 1'
        match = (r".*(Function "+f+r"\("+t+r"\) not supported"
                 r"|Could not bind "+f+r"\("+t+r"\))")
        with pytest.raises(Exception, match=match):
            descr, result = heavydb.sql_execute(q)
            print('query: ', q)
            print('expected: ', match)

    for f in [r'f64', r'f32']:
        for at, av, r in [
                (r'BIGINT', 'i64(0)', (16.5,)),
                (r'INTEGER', 'i32(0)', (12.5,)),
                (r'SMALLINT', 'i_16(0)', (10.5,)),
                (r'TINYINT', 'i_8(0)', (9.5,)),
                (r'BIGINT', 'i8', (8.5,)),
                (r'INTEGER', 'i4', (8.5,)),
                (r'SMALLINT', 'i2', (8.5,)),
                (r'TINYINT', 'i1', (8.5,)),
        ]:
            if available_version[:2] >= (5, 9):
                # heavydb-internal PR 6003 changed the casting table
                if f == r'f32':
                    r = r[0] - 4,
                descr, result = heavydb.sql_execute(
                    f'select {f}({av}) from {heavydb.table_name} limit 1')
                assert list(result)[0] == r, (f, at, av, r)

            elif available_version[:2] == (5, 8):
                # heavydb-internal PR 5814 changed the casting table
                if f == r'f32' and at == r'BIGINT':
                    with pytest.raises(
                            Exception,
                            match=(r".*(Function "+f+r"\("+at+r"\) not supported"
                                   r"|Could not bind "+f+r"\("+at+r"\))")):
                        descr, result = heavydb.sql_execute(
                            f'select {f}({av}) from {heavydb.table_name} limit 1')
                else:
                    if f == r'f32':
                        r = r[0] - 4,
                    descr, result = heavydb.sql_execute(
                        f'select {f}({av}) from {heavydb.table_name} limit 1')
                    assert list(result)[0] == r, (f, at, av, r)

            elif f == r'f64':  # temporary: allow integers as double arguments
                descr, result = heavydb.sql_execute(
                    f'select {f}({av}) from {heavydb.table_name} limit 1')
                assert list(result)[0] == r

            else:
                with pytest.raises(
                        Exception,
                        match=(r".*(Function "+f+r"\("+at+r"\) not supported"
                               r"|Could not bind "+f+r"\("+at+r"\))")):
                    descr, result = heavydb.sql_execute(
                        'select '+f+'('+av+f') from {heavydb.table_name} limit 1')

    for f in [r'i64', r'i32', r'i16', r'i8']:
        for at, av in [
                (r'DOUBLE', 'f64(0E0)'),
                (r'FLOAT', 'f_32(0E0)'),
                (r'DOUBLE', 'f8'),
                (r'FLOAT', 'f4'),
        ]:
            with pytest.raises(
                    Exception,
                    match=(r".*(Function "+f+r"\("+at+r"\) not supported"
                           r"|Could not bind "+f+r"\("+at+r"\))")):
                descr, result = heavydb.sql_execute(
                    f'select {f}({av}) from {heavydb.table_name} limit 1')


def test_truncate_issue(heavydb):
    heavydb.reset()

    @heavydb('int(f64)', 'int(i64)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 64

    @heavydb('int(f32)', 'int(i32)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 32

    @heavydb('int(i16)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 16

    @heavydb('int(i8)')  # noqa: F811
    def bits(x):  # noqa: F811
        return 8

    descr, result = heavydb.sql_execute(
        'select bits(truncate(2016, 1)) from {heavydb.table_name} limit 1'
        .format(**locals()))
    if available_version[:2] >= (5, 8):
        # heavydb-internal PR 5915 changes the casting table
        assert list(result)[0] in [(64,)]
    else:
        assert list(result)[0] in [(16,), (32,)]

    descr, result = heavydb.sql_execute(
        'select bits(truncate(cast(2016.0 as smallint), 1))'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    if available_version[:2] >= (5, 8):
        assert list(result)[0] == (64,)
    else:
        assert list(result)[0] == (16,)

    descr, result = heavydb.sql_execute(
        'select bits(truncate(cast(2016.0 as int), 1))'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    if available_version[:2] >= (5, 8):
        assert list(result)[0] == (64,)
    else:
        assert list(result)[0] == (32,)

    descr, result = heavydb.sql_execute(
        'select bits(truncate(cast(2016.0 as bigint), 1))'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (64,)

    descr, result = heavydb.sql_execute(
        'select bits(truncate(cast(2016.0 as float), 1))'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    if available_version[:2] >= (5, 8):
        assert list(result)[0] == (64,)
    else:
        assert list(result)[0] == (32,)

    descr, result = heavydb.sql_execute(
        'select bits(truncate(cast(2016.0 as double), 1))'
        ' from {heavydb.table_name} limit 1'
        .format(**locals()))
    assert list(result)[0] == (64,)


def test_unregistering(heavydb):
    heavydb.reset()

    @heavydb('i32(i32)')
    def fahrenheit2celsius(f):
        return (f - 32) * 5 / 9

    _, result = heavydb.sql_execute('select fahrenheit2celsius(40)')
    assert list(result)[0] == (4,)

    heavydb.unregister()

    msg = "Undefined function call"
    with pytest.raises(HeavyDBServerError, match=msg):
        heavydb.sql_execute('select fahrenheit2celsius(40)')


def test_format_type(heavydb):
    def test(s, caller=False):
        with heavydb.targets['cpu']:
            with Type.alias(**heavydb.typesystem_aliases):
                typ = Type.fromobject(s)
                if caller:
                    typ = heavydb.caller_signature(typ)
                return heavydb.format_type(typ)

    assert test('int32 x') == 'int32 x'
    assert test('Column<int32 x | name=y>') == 'Column<int32 x | name=y>'
    assert test('Column<int32 | name=y>') == 'Column<int32 y>'
    assert test('Column<int32 x>') == 'Column<int32 x>'
    assert test('Column<int32>') == 'Column<int32>'
    assert test('Column<int32> z') == 'Column<int32> z'
    assert test('Column<int32> | name=z') == 'Column<int32> z'
    assert test('Column<int32> x | name=z') == 'Column<int32> x | name=z'
    assert test('Column<TextEncodingNone>') == 'Column<TextEncodingNone>'
    assert test('Column<Array<int32>>') == 'Column<Array<int32>>'
    assert test('Column<Array<TextEncodingNone>>') == 'Column<Array<TextEncodingNone>>'

    assert test('ColumnList<Array<int32>>') == 'ColumnList<Array<int32>>'

    assert test('OutputColumn<int32>') == 'OutputColumn<int32>'
    assert test('ColumnList<int32>') == 'ColumnList<int32>'

    assert test('UDTF(ColumnList<int32>)') == 'UDTF(ColumnList<int32>)'
    assert test('int32(ColumnList<int32>)') == 'UDTF(ColumnList<int32>)'
    assert (test('UDTF(int32 x, Column<float32> y, OutputColumn<int64> z)')
            == 'UDTF(int32 x, Column<float32> y, OutputColumn<int64> z)')
    assert test('UDTF(RowMultiplier)') == 'UDTF(RowMultiplier)'
    assert test('UDTF(RowMultiplier m)') == 'UDTF(RowMultiplier m)'
    assert test('UDTF(RowMultiplier | name=m)') == 'UDTF(RowMultiplier m)'
    assert test('UDTF(Constant m)') == 'UDTF(Constant m)'
    assert test('UDTF(ConstantParameter m)') == 'UDTF(ConstantParameter m)'
    assert test('UDTF(SpecifiedParameter m)') == 'UDTF(SpecifiedParameter m)'
    assert test('UDTF(PreFlight m)') == 'UDTF(PreFlight m)'
    assert test('UDTF(TableFunctionManager mgr)') == 'UDTF(TableFunctionManager mgr)'
    assert test('UDTF(Cursor<int32, float64>)') == 'UDTF(Cursor<Column<int32>, Column<float64>>)'
    assert test('UDTF(Cursor<int32 x>)') == 'UDTF(Cursor<Column<int32> x>)'
    assert test('UDTF(Cursor<int32 | name=x>)') == 'UDTF(Cursor<Column<int32> x>)'
    assert test('UDTF(Cursor<TextEncodingNone x>)') == 'UDTF(Cursor<Column<TextEncodingNone> x>)'

    assert test('int32(int32)') == '(int32) -> int32'
    assert test('int32(int32 x)') == '(int32 x) -> int32'
    assert test('int32(Array<int32>)') == '(Array<int32>) -> int32'
    assert test('int32(int32[])') == '(Array<int32>) -> int32'
    assert test('int32(Array<int32> x)') == '(Array<int32> x) -> int32'
    assert test('int32(TextEncodingNone)') == '(TextEncodingNone) -> int32'
    assert test('int32(TextEncodingNone x)') == '(TextEncodingNone x) -> int32'
    assert test('int32(TextEncodingDict)') == '(TextEncodingDict) -> int32'
    assert test('int32(TextEncodingDict x)') == '(TextEncodingDict x) -> int32'
    assert test('int32(Array<TextEncodingNone> x)') == '(Array<TextEncodingNone> x) -> int32'

    assert test('int32(RowFunctionManager, int32)') == '(RowFunctionManager, int32) -> int32'

    def test2(s):
        return test(s, caller=True)

    assert test2('UDTF(int32, OutputColumn<int32>)') == '(int32) -> (Column<int32>)'
    assert test2('UDTF(OutputColumn<int32>)') == '(void) -> (Column<int32>)'
    assert (test2('UDTF(int32 | sizer, OutputColumn<int32>)')
            == '(RowMultiplier) -> (Column<int32>)')
    assert (test2('UDTF(ConstantParameter, OutputColumn<int32>)')
            == '(ConstantParameter) -> (Column<int32>)')
    assert test2('UDTF(SpecifiedParameter, OutputColumn<int32>)') == '(void) -> (Column<int32>)'
    assert test2('UDTF(Constant, OutputColumn<int32>)') == '(void) -> (Column<int32>)'
    assert test2('UDTF(PreFlight, OutputColumn<int32>)') == '(void) -> (Column<int32>)'
    assert test2('UDTF(TableFunctionManager, OutputColumn<int32>)') == '(void) -> (Column<int32>)'
    assert (test2('UDTF(RowMultiplier, OutputColumn<Array<int32>>)')
            == '(RowMultiplier) -> (Column<Array<int32>>)')
    assert (test2('UDTF(RowMultiplier, OutputColumn<Array<TextEncodingNone>>)')
            == '(RowMultiplier) -> (Column<Array<TextEncodingNone>>)')

    assert test2('UDTF(Cursor<int32>)') == '(Cursor<Column<int32>>) -> void'
    assert test2('UDTF(Cursor<int32 x>)') == '(Cursor<Column<int32> x>) -> void'


def test_reconnect(heavydb):

    heavydb2 = heavydb.reconnect()

    assert heavydb.session_id != heavydb2.session_id
    assert heavydb2.host == heavydb.host
    assert heavydb2.port == heavydb.port


def test_non_admin_user(heavydb):
    heavydb.require_version((5, 9), 'Requires omniscidb 5.9 or newer')

    user = 'rbc_test_non_admin_user'
    password = 'Xy2kq_3lM'
    dbname = 'rbc_test_non_admin_user_db'

    # create user and user's database:
    heavydb.sql_execute(f'DROP DATABASE IF EXISTS {dbname};')
    heavydb.sql_execute(f'DROP USER IF EXISTS "{user}";')

    heavydb.sql_execute(
        f"CREATE USER {user} (password = '{password}', is_super = 'false', can_login='true');")
    heavydb.sql_execute(f"CREATE DATABASE {dbname} (owner = '{user}');")
    heavydb.sql_execute(f"ALTER USER {user} (default_db = '{dbname}');")

    # test the ability to register and use UDFs as a non-admin user:
    userdb = heavydb.reconnect(user=user, password=password, dbname=dbname)

    assert userdb.user == user
    assert userdb.password == password
    assert userdb.dbname == dbname

    @userdb('int(int)')
    def rbc_test_non_admin_user_udf(x):
        return x + 1

    assert rbc_test_non_admin_user_udf(1).execute() == 2

    # clean up:
    heavydb.sql_execute(f'DROP DATABASE IF EXISTS {dbname};')
    heavydb.sql_execute(f'DROP USER IF EXISTS "{user}";')
